import os
import numpy
import platform
import shutil
from distutils.core import Distribution, Extension

from Cython.Build import build_ext, cythonize

# ========================================
# Build file written by M. Liebsch for the
# fem_c module.
#
# ========================================

# compiler arguments are system dependent
if platform.system() == 'Windows':
    compile_args = ['/O2', '-openmp']
    extra_link_args = []
elif platform.system() == 'Linux':
    compile_args = ['-O3', '-fopenmp']
    extra_link_args = ['-fopenmp']

fem_c_dir = "fem_c"

extensions = [Extension(
    "fem_c_mod",
    [
        os.path.join(fem_c_dir, "interface_c.pyx"),
    ],
    extra_compile_args=compile_args,
    include_dirs=[numpy.get_include()],
    extra_link_args=extra_link_args,
    ),
]

ext_modules = cythonize(extensions, include_path=[fem_c_dir])

dist = Distribution({"ext_modules": ext_modules})
cmd = build_ext(dist)
cmd.ensure_finalized()
cmd.run()

for output in cmd.get_outputs():
    relative_extension = os.path.relpath(output, cmd.build_lib)
    shutil.copyfile(output, os.path.join("snoopy", relative_extension))


import numpy as np
import gmsh

class FiniteElement():
    '''This class provides all functionality for finite elements.
    '''

    def __init__(self, gmsh_code):
        '''Default constructor. It initialized gmsh if this was not done
        already.
        '''
        if not gmsh.isInitialized():
            gmsh.initialize()

        # this is the gmsh code for this type of element
        self.gmsh_code = gmsh_code

        # get the number of nodes
        self.num_nodes = self.get_number_of_nodes()

    def get_number_of_nodes(self):
        '''Get the number of nodes of this element.
        
        :return:
            The number of nodes as integer.
        '''
        # evaluate the basis functions
        _, basis, _ = gmsh.model.mesh.getBasisFunctions(self.gmsh_code, np.array([0., 0., 0.]), "Lagrange")

        # the number of nodes
        return len(basis)

    def evaluate_basis(self, q):
        '''Evaluate the basis functions.
        
        :param q:
            The local coordinates in an array with M*3 elements, where
            M is the number of points. This is the default gmsh style, where
            all quadrature points are always given in u, v, w, i.e. 3D local
            coordinates.

        :return:
            The six basis functions evaluated at the M points.
            i.e. an (M x K) array where K is the number of nodes.
        '''
        
        # this is the number of evaluation points
        M = np.int32(len(q)/3)

        # evaluate the basis functions
        _, basis, _ = gmsh.model.mesh.getBasisFunctions(self.gmsh_code, q, "Lagrange")

        # reshape
        basis.shape = (M, self.num_nodes)

        # return
        return basis

    def evaluate_basis_derivative(self, q):
        '''Evaluate the derivatives of the basis functions.
        
        :param q:
            The local coordinates in an array with M*3 elements, where
            M is the number of points. This is the default gmsh style, where
            all quadrature points are always given in u, v, w, i.e. 3D local
            coordinates.

        :return:
            The eight basis function derivatives evaluated at the M points.
            i.e. an (M x K x 3) array. Where K is the number of nodes.
        '''

        # this is the number of evaluation points
        M = np.int32(len(q)/3)

        # evaluate the basis functions
        _, basis, _ = gmsh.model.mesh.getBasisFunctions(self.gmsh_code, q, "GradLagrange")

        # reshape
        basis.shape = (M, self.num_nodes, 3)

        # return
        return basis

    def evaluate(self, c, nodes, basis):
        '''Evaluate the positions given the evaluated basis
        functions at some local coordinates.
        
        :param c:
            The node connectivity of this element.
        
        :param nodes:
            The nodes.

        :param basis:
            The basis functions evaluated at the local coordinates.

        :return:
            The positions at the M evaluation points in an M x 3 array.
        '''

        # the number of evaluation points
        num_eval = basis.shape[0]

        # return values
        ret_val = np.zeros((num_eval, 3))

        # add basis functions
        for i in range(self.num_nodes):
                        
            ret_val[:, 0] += basis[:, i]*nodes[c[i], 0]
            ret_val[:, 1] += basis[:, i]*nodes[c[i], 1]
            ret_val[:, 2] += basis[:, i]*nodes[c[i], 2]

        return ret_val

    def evaluate_derivative(self, c, nodes, basis_der):
        '''Evaluate the spatial derivative given the evaluated basis
        functions at some local coordinates.
        
        :param c:
            The node connectivity of this element.
        
        :param nodes:
            The nodes.

        :param basis:
            The basis functions evaluated at the local coordinates.

        :return:
            The positions at the M evaluation points in an M x 3 array.
        '''

        # the number of evaluation points
        num_eval = basis_der.shape[0]

        # return values
        ret_val = np.zeros((num_eval, 3, 2))

        # add basis functions
        for i in range(self.num_nodes):

            ret_val[:, 0, 0] += basis_der[:, i, 0]*nodes[c[i], 0]
            ret_val[:, 1, 0] += basis_der[:, i, 0]*nodes[c[i], 1]
            ret_val[:, 2, 0] += basis_der[:, i, 0]*nodes[c[i], 2]

            ret_val[:, 0, 1] += basis_der[:, i, 1]*nodes[c[i], 0]
            ret_val[:, 1, 1] += basis_der[:, i, 1]*nodes[c[i], 1]
            ret_val[:, 2, 1] += basis_der[:, i, 1]*nodes[c[i], 2]

        return ret_val

    def get_quadrature_rule(self, order):
        '''Get the quadrature nodes and weigths for
        this element.
        
        :param order:
            The order of the quadrature rule.
            
        :retrurn:
            The array of M weights and the local coordinates of the points in an 3*M array.
        '''

        q, weights = gmsh.model.mesh.getIntegrationPoints(self.gmsh_code, "Gauss" + str(order))

        return weights, q
        
    def get_edge_basis_functions(self, mesh, loc, element_type, element_order):
        '''Get the edge elements at the integration points.

        :param mesh:
            A gmsh mesh object.
        
        :param loc:
            The integration local coordinates

        :param element_type:
            The type of the finite element (gmsh).

        :param element_order:
            The element order.

        :return:
            A numpy array with all curls of basis functions. Also the number of orientations.
        '''

        # this is the function type string
        function_type = 'HcurlLegendre' + str(element_order-1)

        # evaluate the edge basis functions at the integration points on the reference element
        _, basis_functions, num_orinentations = mesh.getBasisFunctions(element_type, loc.flatten(), function_type)

        return basis_functions, num_orinentations

    def get_curl_edge_basis_functions(self, mesh, loc, element_type, element_order):
        '''Get the curl of the edge elements at the integration points.

        :param mesh:
            A gmsh mesh object.
        
        :param loc:
            The integration local coordinates

        :param element_type:
            The type of the finite element (gmsh).

        :param element_order:
            The element order.

        :return:
            A numpy array with all curls of basis functions. Also the number of orientations.
        '''

        # this is the function type string
        function_type = 'CurlHcurlLegendre' + str(element_order-1)

        # evaluate the edge basis functions at the integration points on the reference element
        _, basis_functions, num_orinentations = mesh.getBasisFunctions(element_type, loc.flatten(), function_type)

        return basis_functions, num_orinentations

    def compute_J(self, element, nodes, d_phi, dim=3):
        '''Compute the Jacobian matrix,
        given the connectivity of a finite element and the nodal coordinates,
        as well as the derivatives of the element basis functions.

        :param element:
            The connectivity of the element.

        :param nodes:
            The (all) nodal coordinates.

        :param d_phi:
            The derivatives of the shape functions of the finite element.

        :param dim:
            The dimension of the finite element. Default = 3.

        :return:
            The Jacobian in an (M x 3 x dim) numpy array.
        '''

        # this is the number of nodes for the finite element
        num_nodes = len(element)

        # this is the number of evaluation points
        num_eval = d_phi.shape[0]

        # this is the return data
        J = np.zeros((num_eval, 3, dim))

        # fill it
        for i in range(num_nodes):
            for j in range(3):
                for k in range(dim):
                    J[:, j, k] += nodes[element[i], j]*d_phi[:, i, k]

        return J

    def compute_J_inv(self, J):
        '''Invert a numpy array of Jacobian matrices.

        :param J:
            A numpy array of Jacobian matrices (M x 3 x 3).

        :return:
            The inverse of these matrices.
        '''

        # the number of evaluation points
        num_eval = J.shape[0]

        # the return data
        inv_J = 0.0*J

        for i in range(num_eval):
            inv_J[i, :, :] = np.linalg.inv(J[i, :, :])

        return inv_J

    def compute_J_det(self, J, dim=3):
        '''Compute the determinants of an array of Jacobian matrices. 

        :param J:
            A numpy array of Jacobian matrices (M x 3 x 3).

        :param dim:
            The dimension of the finite element. Default = 3.

        :return:
            The inverse of these matrices.
        '''

        # the number of evaluation points
        num_eval = J.shape[0]

        # the return data
        det_J = np.zeros((num_eval, ))

        if dim == 3:

            for i in range(num_eval):
                det_J[i] = np.linalg.det(J[i, :, :])

        elif dim == 2:
            # we compute the surface element
            for i in range(num_eval):
                det_J[i] = np.linalg.norm(np.cross(J[i, :, 0], J[i, :, 1]))

        else:
            print('Error! Unknown dimension {}'.format(dim))

        return det_J


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy.interpolate import splev, splrep, PchipInterpolator
from scipy.optimize import minimize_scalar

class Permeability():

    def __init__(self, input_filename):
        '''Default constructor.
        
        :param input_filename:
            The name of the input file with the B(H) data

        :return:
            None.
        '''

        mat_data = pd.read_csv(input_filename)

        self.B_data = mat_data['B(T)'].values
        self.H_data = mat_data['H(A/m)'].values

        self.initialize_interpolation()

        return None

    def initialize_interpolation(self):
        '''Initialize the BSpline spline interpolation.
        
        :return:
            None
        '''

        # allocate mu array
        self.mu = np.zeros((len(self.H_data), ))

        # compute mu
        self.mu[1:] = self.B_data[1:]/self.H_data[1:]
        self.mu[0] = self.mu[1]

        # we fit mu(x) with x = log10(H)
        # the derivative of mu wrt H is
        # dmu_dH = dmu_dx / H / ln(10)
        self.x = np.zeros((len(self.H_data), ))
        self.x[1:] = np.log10(self.H_data[1:])
        self.x[0] = 2*self.x[1] - self.x[2]

        # interpolate
        self.spl = splrep(self.x, self.mu)

        # find the zero derivative in the first interval
        def obj_fcn(x):
            return abs(splev(x, self.spl, der=1))

        # this is the minimum x value we cut off here
        self.x_min = minimize_scalar(obj_fcn, bounds=[self.x[0], self.x[1]]).x

        # this is the corresponding H value
        self.H_min = 10.0**self.x_min

        x_hr = np.linspace(self.x_min, max(self.x), 1000)

        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.plot(self.x[1:], self.mu[1:], 'o', label='data')
        ax.plot(x_hr, splev(x_hr, self.spl), label='fit')
        ax.set_xlabel('$x$')
        ax.set_title(r'$\mu$ in Vs/Am')
        ax.legend()
        plt.show()

        return None

    def evaluate_mu(self, H_mag, relative=False):
        '''Compute the permeability given the field vector magnitude.
        
        :param H_mag:
            The field vector magnitudes.

        :param relative:
            Set this flag to true to return the relative permeability.

        :return:
            The permeability values.
        '''


        # here we avoid zeros
        H = H_mag.copy()
        H[H < self.H_min] = self.H_min

        # the derivative wrt x
        mu = splev(np.log10(H), self.spl)

        if relative:
            return mu/4/np.pi*1e7
        else:
            return mu

    def evaluate_mu_derivative(self, H_mag, relative=False):
        '''Compute the permeability derivative wrt H_mag, given the field vector magnitude.
        
        :param H_mag:
            The field vector magnitudes.

        :param relative:
            Set this flag to true to return the relative permeability.

        :return:
            The permeability derivative values.
        '''


        # here we avoid zeros
        H = H_mag.copy()
        H[H < self.H_min] = self.H_min

        d_mu_dx = splev(np.log10(H), self.spl, der=1)

        # the derivative w.r.t. H
        d_mu = d_mu_dx/H/np.log(10)

        if relative:
            return d_mu/4/np.pi*1e7
        else:
            return d_mu

    def plot_report(self):
        '''Plot the report of fit and data.

        :return:
            None
        '''
        
        # mask out potential zeros
        mask = self.H_data > 0.0

        # make a HR H array
        H_hr = np.logspace(-6, np.log10(max(self.H_data)), 100)

        # compute relative permeability
        mu_r = self.evaluate_mu(H_hr, relative=True)

        # compute also the derivative wrt H
        d_mu_r = self.evaluate_mu_derivative(H_hr, relative=True)

        # validate numerically
        d_mu_r_num = (mu_r - self.evaluate_mu(H_hr-1e-3, relative=True))/1e-3

        # compute permeability
        mu = 4*np.pi*1e-7*mu_r

        fig = plt.figure()
        ax = fig.add_subplot(221)
        ax.plot(self.H_data, self.B_data, 'o', label='data')
        ax.plot(H_hr, H_hr*mu, label='fit')
        ax.set_xlabel('$H$ in A/m')
        ax.set_title('$B$ in T')
        ax.set_xscale('log')
        ax.legend()
        
        ax = fig.add_subplot(222)
        ax.plot(self.H_data[mask], self.B_data[mask]/self.H_data[mask]/4/np.pi*1e7, 'o', label='data')
        ax.plot(H_hr, mu_r, label='fit')
        ax.set_xlabel('$H$ in A/m')
        ax.set_title(r'$\mu_{\text{r}}$ in Vs/Am')
        ax.legend()
        ax.set_xscale('log')

        
        ax = fig.add_subplot(223)
        ax.plot(H_hr, d_mu_r, label='analytical')
        ax.plot(H_hr, d_mu_r_num, '--', label='numerical')
        ax.set_xlabel('$H$ in A/m')
        ax.set_title(r'$\text{d}\mu_{\text{r}}/\text{d} H$ in Vs/A$^2$')
        ax.legend()
        ax.set_xscale('log')

        plt.show()
        
    def get_spline_information(self):
        '''Get the knots t, the control points c and the degree k of this spline.
        Get also the lower limit of the spline interpolation (H_min) and the
        permeability for H < H_min.
        
        :return:
            Knots vector t,
            Control points c,
            Spline degree k,
            The minimum H value,
            The value of mu for H < H_min.
        '''

        return self.spl[0], self.spl[1], self.spl[2], self.H_min, self.evaluate_mu(np.array([0.0]))[0]

    def get_type_spec(self):
        '''Return the type specifyer. 
        0 ConstantPermeability 
        1 Permeability (BSpline)
        
        :return:
            The type specifyer.
        '''

        return 1

class ConstantPermeability():

    def __init__(self, mu):
        '''Default constructor.
        
        :param mu:
            The value for the constant permeability.

        :return:
            None.
        '''

        self.mu = mu

        return None

    def evaluate_mu(self, H_mag, relative=False):
        '''Compute the permeability given the field vector magnitude.
        
        :param H_mag:
            The field vector magnitudes.

        :param relative:
            Set this flag to true to return the relative permeability.

        :return:
            The permeability values.
        '''

        if relative:
            return self.mu/4/np.pi*1e7 + 0.0*H_mag
        else:
            return self.mu + 0.0*H_mag

    def evaluate_mu_derivative(self, H_mag, relative=False):
        '''Compute the permeability derivative wrt H_mag, given the field vector magnitude.
        
        :param H_mag:
            The field vector magnitudes.

        :param relative:
            Set this flag to true to return the relative permeability.

        :return:
            The permeability derivative values.
        '''

        return 0.0 + 0.0*H_mag

    def get_type_spec(self):
        '''Return the type specifyer. 
        0 ConstantPermeability 
        1 Permeability (BSpline)
        
        :return:
            The type specifyer.
        '''

        return 0

class Reluctance():

    def __init__(self, input_filename):
        '''Default constructor.
        
        :param input_filename:
            The name of the input file with the B(H) data

        :return:
            None.
        '''

        mat_data = pd.read_csv(input_filename)

        self.B_data = mat_data['B(T)'].values
        self.H_data = mat_data['H(A/m)'].values

        self.initialize_interpolation()

        return None

    def initialize_interpolation(self):
        '''Initialize the BSpline spline interpolation.
        
        :return:
            None
        '''

        # compute nu
        self.nu = 0.0*self.H_data
        self.nu[1:] = self.H_data[1:]/self.B_data[1:]
        self.nu[0] = self.nu[1]

        # interpolate
        self.spl = PchipInterpolator(self.B_data, self.nu, extrapolate=True)
        # self.spl = splrep(self.B_data, self.nu)

        # B_hr = np.linspace(min(self.B_data), 1.5*max(self.B_data), 100)


        # nu_alt = self.evaluate_nu_alt(B_hr)

        # dnu = self.evaluate_nu_derivative(B_hr)
        # dnu_alt = self.evaluate_nu_derivative_alt(B_hr)

        # fig = plt.figure()
        # ax = fig.add_subplot(121)
        # ax.plot(self.B_data, self.nu, 'o', label='data')
        # ax.plot(B_hr, self.spl(B_hr), label='fit')
        # ax.plot(B_hr, nu_alt, '--', label='fit alt')
        # ax.set_xlabel('$B$ in T')
        # ax.set_title(r'$\nu$ in Am/Vs')
        # ax.legend()
        # ax = fig.add_subplot(122)
        # ax.plot(B_hr, dnu, label='python')
        # ax.plot(B_hr, dnu_alt, '--', label='my')
        # ax.set_xlabel('$B$ in T')
        # ax.set_title(r'$\nu$ derivative in Am/TVs')
        # ax.legend()
        # plt.show()

        return None

    def evaluate_nu(self, B_mag):
        '''Compute the reluctance given the B field vector magnitude.
        
        :param B_mag:
            The B field vector magnitudes.

        :return:
            The reluctance values.
        '''

        return self.spl(B_mag)
    
    def evaluate_nu_alt(self, B_mag):
        '''Compute the reluctance given the B field vector magnitude.
        
        :param B_mag:
            The B field vector magnitudes.

        :return:
            The reluctance values.
        '''
        c = self.spl.c
        x = self.spl.x

        nu = 0.0*B_mag

        for j in range(len(B_mag)):
            for i in range(len(x) - 1):
                if (x[i] <= B_mag[j]) and (B_mag[j] < x[i+1]):
                    for l in range(4):
                        nu[j] += c[l, i]*(B_mag[j] - x[i])**(3-l)

            if (x[-1] <= B_mag[j]):
                for l in range(4):

                    nu[j] += c[l, -1]*(B_mag[j] - x[-2])**(3-l)
        return nu
    
    def evaluate_nu_derivative(self, B_mag):
        '''Compute the reluctance derivative wrt B_mag, given the B field vector magnitude.
        
        :param B_mag:
            The B field vector magnitudes.

        :return:
            The reluctance derivative values.
        '''

        return self.spl.derivative()(B_mag)
    
    def evaluate_nu_derivative_alt(self, B_mag):
        '''Compute the reluctance derivative given the B field vector magnitude.
        This function performs the calculation to evaluate the fit directly.

        :param B_mag:
            The B field vector magnitudes.

        :return:
            The reluctance values.
        '''
        c = self.spl.c
        x = self.spl.x

        dnu = 0.0*B_mag

        for j in range(len(B_mag)):
            for i in range(len(x) - 1):
                if (x[i] < B_mag[j]) and (B_mag[j] <= x[i+1]):
                    for l in range(4):
                        dnu[j] += (3-l)*c[l, i]*(B_mag[j] - x[i])**(2-l)

            if (x[-1] < B_mag[j]):
                for l in range(4):

                    dnu[j] += (3-l)*c[l, -1]*(B_mag[j] - x[-2])**(2-l)
        return dnu
    
    def get_type_spec(self):
        '''Return the type specifyer. 
        0 ConstantReluctance 
        1 Reluctance (BSpline)
        
        :return:
            The type specifyer.
        '''

        return 1

    def get_spline_information(self):
        '''Get the knots t, the control points c of this spline.
        
        :return:
            The spline coefficients
            The B intervals
        '''
        
        return self.spl.c, self.spl.x


    def plot_permeability(self, ax, B_max, resol=100, relative=True, label='none'):
        '''Plot the permeability over B.

        :param ax:
            The matplotlib axes to plot into.

        :param B_max:
            The maximum B field for the plot.

        :param resol:
            The resolution. Default 100.

        :param relative:
            Set this flag true to plot the relative permeability.

        :param label:
            A plot label. If 'none' no label is used.

        :return:
            None
        '''

        # the B field values
        B = np.linspace(0, B_max, resol)

        # the reluctance
        nu = self.evaluate_nu(B)

        # the permeabiltiy
        mu = 1.0/nu

        if relative:

            if label == 'none':
                ax.plot(B, mu/4/np.pi*1e7)
            else:
                ax.plot(B, mu/4/np.pi*1e7, label=label)

        else:

            if label == 'none':
                ax.plot(B, mu)
            else:
                ax.plot(B, mu, label=label)

        return

class ConstantReluctance():

    def __init__(self, value):
        '''Default constructor.
        
        :param value:
            The reluctance value.

        :return:
            None.
        '''

        self.value = value

        return None

    def evaluate_nu(self, B_mag):
        '''Compute the reluctance given the B field vector magnitude.
        
        :param B_mag:
            The B field vector magnitudes.

        :return:
            The reluctance values.
        '''

        return self.value + 0.0*B_mag

    def evaluate_nu_derivative(self, B_mag):
        '''Compute the reluctance derivative wrt B_mag, given the B field vector magnitude.
        
        :param B_mag:
            The B field vector magnitudes.

        :return:
            The reluctance derivative values.
        '''

        return 0.0*B_mag
    
    def get_type_spec(self):
        '''Return the type specifyer. 
        0 ConstantReluctance 
        1 Reluctance (BSpline)
        
        :return:
            The type specifyer.
        '''

        return 0


import numpy as np
import gmsh
import pyvista as pv

from scipy.sparse.linalg import spsolve
from scipy.sparse.linalg import cg
from scipy.sparse import diags

import matplotlib.pyplot as plt

from .laplace_subdomain_solver import LaplaceSubdomainSolver
from .curl_curl_assembler import CurlCurlAssembler
from .matrix_assembly import delete_rows_sparse
from .matrix_assembly import delete_cols_sparse
from .matrix_assembly import recover_full_solution

from .materials import ConstantReluctance

from .mesh_tools import get_edge_boundary_dofs
from .mesh_tools import get_cotree_dofs

from .plot_tools import plot_vector_field

class RedMVPSolver():

    def __init__(self, gmsh_model, coil_list, iron_tags,
                 air_tags, reluctances,
                 max_newton_iterations=25,
                 element_order=1,
                 quad_order=6):
        '''Setup the solver with the geometry info.
        
        :param gmsh_model:
            The gmsh model object.
            
        :param coil_list:
            A list of coil objects. They need to have a function 'compute_B(points)'
            for the field evaluation.

        :param iron_tags:
            The tags for the iron domains.

        :param air_tags:
            The tags for the air domains.

        :param reluctances:
            A list with reluctances for the iron domains. In the same order
            than the iron_tags.
        :param element_order:
            The order of the FEM approximation space for the vector potential.

        :param max_newton_iteration:
            The maximum number of Newton iterations.

        :return:
            None
        '''

        # initialize gmsh if not done already
        if not gmsh.isInitialized():
            gmsh.initialize()

        # store the tags for the iron domains
        self.iron_tags = iron_tags

        # store the tags for the iron domains
        self.air_tags = air_tags

        # store the excitation coils
        self.coil_list = coil_list

        # append all domain tags for the vector potential solver
        domain_tags = iron_tags + air_tags

        # append all material properties
        material_properties = reluctances + [ConstantReluctance(0.25/np.pi*1e7)]*len(air_tags)

        # the maximum number of Newton iterations
        self.max_newton_iterations = max_newton_iterations

        # store the finite element order
        self.element_order = element_order

        if (element_order > 3):
            print('Error! Element order {} not implemented! Choosing 1.')
            element_order = 1

        self.quad_order = quad_order

        # get the dirichlet boundary dofs
        self.dirichlet_dofs = get_edge_boundary_dofs(gmsh_model, ["Dirichlet Boundary"], element_order)

        # setup the matrix factory for the vector potential solver
        self.curl_curl_factory = CurlCurlAssembler(gmsh_model.mesh, domain_tags, material_properties, element_order)

        # the number of dofs for the vector potential
        self.num_dofs_mvp = self.curl_curl_factory.num_dofs


    def get_global_ids(self):
        '''Get the global ids for the field solution.

        :return:
            A (E x N) matrix (int) where E is the number of elements and N is the number
            of edge basis functions per element. 
        '''
        return self.curl_curl_factory.global_ids

    def make_boundary_mask(self):
        '''Make a mask to mask out the boundary degrees of freedom.

        :return:
            The boundary mask.
        '''

        # make the mask
        mask = np.ones((self.num_dofs_mvp, ), dtype=np.bool)

        # set the boundary dofs to fals
        for bc in self.dirichlet_dofs:
            mask[bc-1] = False

        return mask

    def solve(self, tolerance = 1e-4, x_0=np.zeros((0, )), maxiter=-1):
            '''Solve the problem.

            :param tolerance:
                The tolerance for the cg iterations.

            :param xcg_0:
                The initial guess for the cg iterations.
                
            :param maxiter:
                The maximum number of iterations. If -1, the default scipy setting is used.

            :param apply_gauge:
                Set this flag true to enable the tree-cotree gauged formulation.

            :return:
                The solution vector.
            '''


            # get the number of iron and air domains for convinience
            num_iron_domains = len(self.iron_tags)
            num_air_domains = len(self.air_tags)
            num_domains = num_iron_domains+num_air_domains

            # the initial solution vector for the mvp here 0, a linear problem
            if len(x_0) == 0:
                x_0 = np.zeros((self.num_dofs_mvp, ))

            # compute the source fields
            # we only need to compute them in the iron domains
            quad_points = self.curl_curl_factory.get_quadrature_points(self.quad_order, [i for i in range(num_iron_domains)])

            # we evaluate the source field
            source_fields = []

            for i in range(num_iron_domains):
                B_src = 0.0*quad_points[i]
                for coil in self.coil_list:
                    B_src += coil.compute_B(quad_points[i])
                source_fields.append(B_src)

            for i in range(num_air_domains):
                source_fields.append(np.zeros((0, 3)))

            # make a boundary mask for the dirichlet boundary condition
            mask = self.make_boundary_mask()

            # initalize current solution vector
            x_n = x_0.copy()

            print('***********************')
            print('start Newton iterations')
            for i in range(self.max_newton_iterations):
                

                # compute the stiffness and Jacobi matrix
                K, J, rhs = self.curl_curl_factory.compute_stiffness_and_jacobi_matrix_c(x_n,
                                                                            quad_order=self.quad_order,
                                                                            source_fields=source_fields)

                # compute right hand side
                b =  J @ x_n - K @ x_n + rhs

                # check if zero excitation
                if max(abs(b)) < 1e-14:
                    x_np1 = 0.0*x_n
                    print('  zero excitation!')
                    break

                # reduce equation system (impose boundary conditions)
                J_c = J[mask, :]
                J_c = J_c[:, mask]

                # jacobi preconditioner
                M = diags(1.0/J_c.diagonal(), shape=J_c.shape)

                
                # launch cg
                if maxiter < 0:
                    xx, exit_code = cg(J_c, b[mask], atol=tolerance, x0=x_n[mask], M=M)
                else:
                    xx, exit_code = cg(J_c, b[mask], atol=tolerance, x0=x_n[mask], M=M, maxiter=maxiter)


                # we need to get back to the full solution vector
                x_np1 = 0.0*x_n
                x_np1[mask == True] = xx



                # the step
                h = x_np1 - x_n

                print('  step {}'.format(i))
                print('    max relative increment {:.3e}'.format(max(abs(h))/max(abs(x_np1))))

                # stop if tolerance is reached
                if max(abs(h)) < tolerance*max(abs(x_np1)):
                    print('  tolerance reached!')
                    break
                else:
                    x_n = x_np1.copy()

            return x_np1



import numpy as np
import gmsh
from tqdm import tqdm
from scipy.sparse import csr_array
import matplotlib.pyplot as plt
import pandas as pd

from . import fem_c_mod as fem_c
from .finite_element import FiniteElement
from .mesh_tools import get_vector_basis_mesh_info
from .mesh_tools import get_vector_basis_orientations
from .mesh_tools import get_mesh_info
from .mesh_tools import get_global_ids
from .mesh_tools import get_num_edge_dofs

def recover_full_solution(u, u_g, b):
    '''Given the fem solution and the lifting vector,
    as well as the boundary node list, recover the 
    full FEM solution vector (scalar).

    :param u:
        The FEM solution vector (reduced).
    
    :param u_g:
        The lifting vector.

    :return:
        The full FEM solution.
    '''
    # the total number of nodes
    num_nodes = len(u_g)

    # the indices of the remaining nodes.
    indx_rem = np.linspace(0, num_nodes-1, num_nodes, dtype=np.int32)

    # delete the boundary nodes
    indx_rem = np.delete(indx_rem, b)

    # copy the lifting vector
    u_ret = u_g.copy()

    # fill it
    u_ret[indx_rem] = u

    # return
    return u_ret

def delete_rows_sparse(M, rows):
    '''Delete some rows of a sparse matrix.

    :param M:
        The sparse matrix.

    :param rows:
        The rows to delete.

    :return:
        The resulting sparse matrix.
    '''

    # this is the number of rows of the matrix
    num_rows = M.shape[0]

    mask = np.ones(num_rows, dtype=bool)
    mask[rows] = False

    return M[np.flatnonzero(mask), :]


def delete_cols_sparse(M, cols):
    '''Delete some columns of a sparse matrix.

    :param M:
        The sparse matrix.

    :param cols:
        The columns to delete.

    :return:
        The resulting sparse matrix.
    '''

    # this is the number of cols of the matrix
    num_cols = M.shape[1]

    mask = np.ones(num_cols, dtype=bool)
    mask[cols] = False

    return M[:, np.flatnonzero(mask)]


class GradGradAssembler():

    def __init__(self, mesh, domain_tags):
        '''Default constructor.
        
        :param mesh:
            A gmsh mesh object.

        :param domain_tags:
            The tags for the domains.

        :return:
            None.
        '''

        # The nodes are not sorted correctly. I don't know why...
        # But we need to get them like this:
        node_tags, _, _ = mesh.getNodes()
        num_nodes = len(node_tags)
        node_tags = np.unique(node_tags)

        # we now make an array of unique mesh nodes.
        self.nodes = np.zeros((num_nodes, 3))

        for i in range(num_nodes):
            self.nodes[i, :] = mesh.getNode(node_tags[i])[0]
        
        # we allocate lists for the cell connectivities and types
        # for all materials
        self.num_materials = len(domain_tags)
        self.cell_types = []
        self.cells = []
        
        for i in domain_tags:
            
            # get the elements
            c_types_tmp, _, cells_tmp = gmsh.model.mesh.getElements(3, i)

            # append to the list
            self.cell_types.append(c_types_tmp)
            self.cells.append(cells_tmp)
            
        # the number of nodes
        num_nodes = self.nodes.shape[0]


        # cell_mat = self.cells[0] - 1
        # cell_mat = np.array(cell_mat, dtype=np.int32)
        # cell_mat.shape = (np.int32(len(self.cells[0])/4), 4)

        # edges = [[0, 1],
        #          [1, 2],
        #          [2, 0],
        #          [0, 3],
        #          [1, 3],
        #          [3, 3]]

        # fig = plt.figure()
        # ax = fig.add_subplot(111, projection='3d')
        # for i, c in enumerate(cell_mat):

        #     for e in edges:
        #         ax.plot([self.nodes[c[e[0]], 0], self.nodes[c[e[1]], 0]],
        #                 [self.nodes[c[e[0]], 1], self.nodes[c[e[1]], 1]],
        #                 [self.nodes[c[e[0]], 2], self.nodes[c[e[1]], 2]], '--o' ,color='k')
                
        # plt.show()

        return None
    

    def setup_mesh(self, nodes, cells, cell_types):
        """Setup the mesh from external information.

        :param nodes:
            The nodal coordinates.

        :param cells:
            The mesh connectivity list.

        :param cell_types.
            The mesh cell types list.

        :return:
            None.
        """

        self.nodes = nodes
        self.cells = cells
        self.cell_types = cell_types

        return None
    
    def get_num_dofs(self):
        '''Get the total number of degrees of freedom.
        
        :return:
            An integer specifying the number if DoFs.
        '''
        return self.nodes.shape[0]
    
    def get_quadrature_rule(self, gmsh_code, order):
        '''Get the quadrature points and nodes for a
        finite element of certain type.

        :param gmsh_code:
            The cell type gmsh code.

        :param order:
            The quadrature order.

        :return:
            The points q and the weights w.
        '''
        q, w = gmsh.model.mesh.getIntegrationPoints(gmsh_code, "Gauss" + str(order))
        
        return q, w
    
    def compute_grad_phi(self, c, x, d_phi, inv_J):
        '''Compute the magnetic field in a finite element
        with connectivity c. The vector x stores the nodal 
        values of the solution.
        The array grad_phi stores the gradients of the FEM
        basis functions.

        :param c:
            The node connectivity of the finite element.

        :param x:
            The solution vector.

        :param d_phi:
            The derivatives of all basis functions.

        :param inv_J:
            The inverse of the Jacobian matrix of the transformation.

        :return:
            The field vectors H, as well as the magnitude H_mag.
        '''

        # the number of evaluation points
        num_eval = d_phi.shape[0]

        # allocate the return array
        grad_phi = np.zeros((num_eval, 3))

        # evaluate
        grad_phi[:, 0] = d_phi[:, :, 0] @ x[c]
        grad_phi[:, 1] = d_phi[:, :, 1] @ x[c]
        grad_phi[:, 2] = d_phi[:, :, 2] @ x[c]


        # transform it to the global domain
        for m in range(num_eval):
            grad_phi[m, :] = inv_J[m, :, :].T @ grad_phi[m, :]


        # return also the magnitude
        mag = np.linalg.norm(grad_phi, axis=1)

        return grad_phi, mag
        
    def compute_stiffness_matrix(self, x, mu_list, quad_order=8):
        '''Compute the stiffness matrix for the magnetic scalar potential for
        nonlinear problems.
        This implementation is just to test the equations. We do not focus
        on performance at this point. A fast version of this code will be
        implemented in an optimized C++ code.

        :param x:
            The solution vector.
            
        :param mu_list:
            The nonlinear permeability functions for each domain in a list.

        :param quad_order:
            The quadrature order. Default 8.

        :return:
            The stiffness matrix.
        '''

        print('assembling the stiffness matrix (scalar potential) (py)...')

        # the total number of nodes
        num_nodes = self.nodes.shape[0]

        # number of cell types
        #num_cell_types = len(self.cell_types)

        # make the sparse stiffness matrix
        K = csr_array((num_nodes, num_nodes))
    
        # loop over all materials
        for n in range(self.num_materials):

            # get the geometry info of this volume
            cell_types = self.cell_types[n]
            cells = self.cells[n]

            # get also the permeability in this domain
            mu_fcn = mu_list[n]
            
            # loop over cell types
            for i, ct in enumerate(cell_types):

                # make the finite element
                finite_element = FiniteElement(ct)

                # the numebr fo element nodes
                el_nodes = finite_element.get_number_of_nodes()

                # the number of elements of this type
                num_el = np.int32(len(cells[i]) / el_nodes)

                # get the quadrature nodes and weights
                q, w = self.get_quadrature_rule(ct, quad_order)

                # the number of quadrature points
                num_quad_pts = np.int32(len(q) / 3)

                # evaluate the shape functions at these points (not needed)
                # phi = finite_element.evaluate_basis(q)

                # evaluate the derivatives at these points
                d_phi = finite_element.evaluate_basis_derivative(q)

                # the number of triplets
                num_triplets = num_el*el_nodes*el_nodes

                # allocate the space for the triplet list
                ij = np.zeros((num_triplets, 2), dtype=np.int32)
                vals = np.zeros((num_triplets, ))

                # a triplet counter
                t_cnt = 0

                # loop over all elements
                for e_i in tqdm(range(num_el)):

                    # gmsh starts counting at 1
                    e = cells[i][e_i*el_nodes:(e_i+1)*el_nodes] - 1

                    # compute the Jacobian at the integration points
                    # inv_J is a numpy array of dimension (M x 3 x 3) where
                    # the inverse of the Jacobians are stored in the second and
                    # third dimensions
                    # M is the number of integration points
                    # det_J is an (M x 0) numpy array which stores the M values of det(J)
                    J = finite_element.compute_J(e, self.nodes, d_phi)
                    
                    inv_J = finite_element.compute_J_inv(J)
                    det_J = finite_element.compute_J_det(J)


                    _, H_mag = self.compute_grad_phi(e, x, d_phi, inv_J)

                    # compute the permebility
                    mu = mu_fcn(H_mag)

                    # loop over all combinations of basis functions
                    for j in range(el_nodes):
                        for k in range(j, el_nodes):

                            # this is the integration value
                            int_val = 0.0

                            # gaussian integration
                            for m in range(num_quad_pts):

                                # transform the gradients
                                grad_u = inv_J[m, :, :].T @ d_phi[m, j, :]
                                grad_v = inv_J[m, :, :].T @ d_phi[m, k, :]

                                # increment the integration value
                                int_val += mu[m]*det_J[m]*w[m]*np.sum(grad_u*grad_v)

                            # sparse matrix version

                            # fill the triplet list
                            ij[t_cnt, 0] = e[j]
                            ij[t_cnt, 1] = e[k]
                            vals[t_cnt] = int_val
                            t_cnt += 1

                            # apply symmetry
                            if j != k:
                                ij[t_cnt, 0] = e[k]
                                ij[t_cnt, 1] = e[j]
                                vals[t_cnt] = int_val
                                t_cnt += 1

                            # dense matrix version

                            # fill the matrix
                            # K_dense[e[j], e[k]] += int_val

                            # apply symmetry
                            # if j != k:
                            #     K_dense[e[k], e[j]] += int_val

                # print('number of triplets = {}'.format(len(vals)))

                # make the sparse matrix
                K += csr_array((vals, (ij[:, 0], ij[:, 1])), shape=(num_nodes, num_nodes))

        return K
    

    def compute_stiffness_matrix_red(self, H, mu_list, quad_order=8):
        '''Compute the stiffness matrix for the magnetic scalar potential for
        nonlinear problems.
        This implementation is just to test the equations. We do not focus
        on performance at this point. A fast version of this code will be
        implemented in an optimized C++ code.

        :param H:
            The impressed field.
            
        :param mu_list:
            The nonlinear permeability functions for each domain in a list.

        :param quad_order:
            The quadrature order. Default 8.

        :return:
            The stiffness matrix.
        '''

        print('assembling the stiffness matrix (scalar potential) (py)...')

        # the total number of nodes
        num_nodes = self.nodes.shape[0]

        # number of cell types
        #num_cell_types = len(self.cell_types)

        # make the sparse stiffness matrix
        K = csr_array((num_nodes, num_nodes))
    
        # a global element counter (all materials)
        glob_el_cnt = 0

        # loop over all materials
        for n in range(self.num_materials):

            # get the geometry info of this volume
            cell_types = self.cell_types[n]
            cells = self.cells[n]

            # get also the permeability in this domain
            mu_fcn = mu_list[n]
            
            # loop over cell types
            for i, ct in enumerate(cell_types):

                # make the finite element
                finite_element = FiniteElement(ct)

                # the numebr fo element nodes
                el_nodes = finite_element.get_number_of_nodes()

                # the number of elements of this type
                num_el = np.int32(len(cells[i]) / el_nodes)

                # get the quadrature nodes and weights
                q, w = self.get_quadrature_rule(ct, quad_order)

                # the number of quadrature points
                num_quad_pts = np.int32(len(q) / 3)

                # evaluate the shape functions at these points (not needed)
                # phi = finite_element.evaluate_basis(q)

                # evaluate the derivatives at these points
                d_phi = finite_element.evaluate_basis_derivative(q)

                # the number of triplets
                num_triplets = num_el*el_nodes*el_nodes

                # allocate the space for the triplet list
                ij = np.zeros((num_triplets, 2), dtype=np.int32)
                vals = np.zeros((num_triplets, ))

                # a triplet counter
                t_cnt = 0

                # loop over all elements
                for e_i in tqdm(range(num_el)):

                    # gmsh starts counting at 1
                    e = cells[i][e_i*el_nodes:(e_i+1)*el_nodes] - 1

                    # compute the Jacobian at the integration points
                    # inv_J is a numpy array of dimension (M x 3 x 3) where
                    # the inverse of the Jacobians are stored in the second and
                    # third dimensions
                    # M is the number of integration points
                    # det_J is an (M x 0) numpy array which stores the M values of det(J)
                    J = finite_element.compute_J(e, self.nodes, d_phi)
                    
                    inv_J = finite_element.compute_J_inv(J)
                    det_J = finite_element.compute_J_det(J)

                    # get the field vectors
                    this_H = H[glob_el_cnt*num_quad_pts:(glob_el_cnt+1)*num_quad_pts, :]

                    # compute also the field vector magnitude !!!!!!!!
                    H_mag = np.linalg.norm(this_H, axis=1)

                    # compute the permebility
                    mu = mu_fcn(H_mag)

                    # loop over all combinations of basis functions
                    for j in range(el_nodes):
                        for k in range(j, el_nodes):

                            # this is the integration value
                            int_val = 0.0

                            # gaussian integration
                            for m in range(num_quad_pts):

                                # transform the gradients
                                grad_u = inv_J[m, :, :].T @ d_phi[m, j, :]
                                grad_v = inv_J[m, :, :].T @ d_phi[m, k, :]

                                # increment the integration value
                                int_val += mu[m]*det_J[m]*w[m]*np.sum(grad_u*grad_v)

                            # sparse matrix version

                            # fill the triplet list
                            ij[t_cnt, 0] = e[j]
                            ij[t_cnt, 1] = e[k]
                            vals[t_cnt] = int_val
                            t_cnt += 1

                            # apply symmetry
                            if j != k:
                                ij[t_cnt, 0] = e[k]
                                ij[t_cnt, 1] = e[j]
                                vals[t_cnt] = int_val
                                t_cnt += 1

                            # dense matrix version

                            # fill the matrix
                            # K_dense[e[j], e[k]] += int_val

                            # apply symmetry
                            # if j != k:
                            #     K_dense[e[k], e[j]] += int_val

                # print('number of triplets = {}'.format(len(vals)))

                # make the sparse matrix
                K += csr_array((vals, (ij[:, 0], ij[:, 1])), shape=(num_nodes, num_nodes))

        return K
    
    def compute_stiffness_and_jacobi_matrix(self, x, mu_list, quad_order=8):
        '''Compute the stiffness and the jacobi matrix for the magnetic scalar potential for
        nonlinear problems.
        This implementation is just to test the equations. We do not focus
        on performance at this point. A fast version of this code will be
        implemented in an optimized C++ code.

        :param x:
            The solution vector.
            
        :param mu_list:
            A list of permeability classes for each material.

        :param quad_order:
            The quadrature order. Default 8.

        :return:
            The stiffness matrix.
        '''

        print('assembling the stiffness and jacobi matrix (scalar potential) (py)...')

        # the total number of nodes
        num_nodes = self.nodes.shape[0]

        # number of cell types
        #num_cell_types = len(self.cell_types)

        # make the sparse stiffness matrix
        K = csr_array((num_nodes, num_nodes))
    
        # make the sparse Jacobi matrix
        dK = csr_array((num_nodes, num_nodes))

        # loop over all materials
        for n in range(self.num_materials):

            # get the geometry info of this volume
            cell_types = self.cell_types[n]
            cells = self.cells[n]

            # get also the permeability in this domain
            permeability = mu_list[n]

            # loop over cell types
            for i, ct in enumerate(cell_types):

                # make the finite element
                finite_element = FiniteElement(ct)

                # the numebr fo element nodes
                el_nodes = finite_element.get_number_of_nodes()

                # the number of elements of this type
                num_el = np.int32(len(cells[i]) / el_nodes)

                # get the quadrature nodes and weights
                q, w = self.get_quadrature_rule(ct, quad_order)

                # the number of quadrature points
                num_quad_pts = np.int32(len(q) / 3)

                # evaluate the shape functions at these points (not needed)
                # phi = finite_element.evaluate_basis(q)

                # evaluate the derivatives at these points
                d_phi = finite_element.evaluate_basis_derivative(q)

                # the number of triplets
                num_triplets = num_el*el_nodes*el_nodes

                # allocate the space for the triplet list (K)
                ij = np.zeros((num_triplets, 2), dtype=np.int32)
                vals_K = np.zeros((num_triplets, ))

                # allocate the space for the triplet list (J)
                vals_dK = np.zeros((num_triplets, ))

                # a triplet counter
                t_cnt = 0

                # loop over all elements
                for e_i in tqdm(range(num_el)):

                    # gmsh starts counting at 1
                    e = cells[i][e_i*el_nodes:(e_i+1)*el_nodes] - 1

                    # compute the Jacobian at the integration points
                    # inv_J is a numpy array of dimension (M x 3 x 3) where
                    # the inverse of the Jacobians are stored in the second and
                    # third dimensions
                    # M is the number of integration points
                    # det_J is an (M x 0) numpy array which stores the M values of det(J)
                    J = finite_element.compute_J(e, self.nodes, d_phi)
                    
                    inv_J = finite_element.compute_J_inv(J)
                    det_J = finite_element.compute_J_det(J)

                    # compute also the field vector
                    grad_phi, H_mag = self.compute_grad_phi(e, x, d_phi, inv_J)

                    # compute the permebility
                    mu = permeability.evaluate_mu(H_mag)


                    # compute the permebility derivative
                    d_mu = permeability.evaluate_mu_derivative(H_mag)

                    # loop over all combinations of basis functions
                    for j in range(el_nodes):
                        for k in range(j, el_nodes):

                            # this is the integration value (K)
                            int_val_K = 0.0

                            # this is the integration value (dK)
                            int_val_dK = 0.0       


                            # gaussian integration
                            for m in range(num_quad_pts):

                                # transform the gradients
                                grad_u = inv_J[m, :, :].T @ d_phi[m, j, :]
                                grad_v = inv_J[m, :, :].T @ d_phi[m, k, :]

                                # increment the integration value
                                int_val_K += mu[m]*det_J[m]*w[m]*np.sum(grad_u*grad_v)

                                # increment the integration value
                                if H_mag[m] >= 1e-14:
                                    
                                    int_val_dK += d_mu[m]*det_J[m]*w[m]*np.sum(grad_u*grad_phi[m, :])*np.sum(grad_v*grad_phi[m, :])/H_mag[m]
                                    
                            # sparse matrix version

                            # fill the triplet list
                            ij[t_cnt, 0] = e[j]
                            ij[t_cnt, 1] = e[k]
                            vals_K[t_cnt] = int_val_K
                            vals_dK[t_cnt] = int_val_dK
                            t_cnt += 1

                            # apply symmetry
                            if j != k:
                                ij[t_cnt, 0] = e[k]
                                ij[t_cnt, 1] = e[j]
                                vals_K[t_cnt] = int_val_K
                                vals_dK[t_cnt] = int_val_dK
                                t_cnt += 1

                            # dense matrix version

                            # fill the matrix
                            # K_dense[e[j], e[k]] += int_val

                            # apply symmetry
                            # if j != k:
                            #     K_dense[e[k], e[j]] += int_val

                # print('number of triplets = {}'.format(len(vals)))

                # make the sparse matrix
                K += csr_array((vals_K, (ij[:, 0], ij[:, 1])), shape=(num_nodes, num_nodes))
                dK += csr_array((vals_dK, (ij[:, 0], ij[:, 1])), shape=(num_nodes, num_nodes))

            # add K and dK
            dK += K

        return K, dK

    def compute_rhs_and_jacobi_matrix(self, H, mu_list, quad_order=8):
        '''Compute the rhs and the jacobi matrix for the magnetic scalar potential for
        nonlinear problems. The H field vectors may include exterior field so that 
        this function can be used for the reduced vector potential formulation.
        
        This implementation is just to test the equations. We do not focus
        on performance at this point. A fast version of this code will be
        implemented in an optimized C++ code.

        :param x:
            The solution vector.
            
        :param mu_list:
            A list of permeability classes for each material.

        :param quad_order:
            The quadrature order. Default 8.

        :param H_ext:
            An external field for the reduced scalar potential formulation.
            If this array has zero rows (default), the external field is ignored.
            Otherwise, the number of rows must match the number of integration points.

        :return:
            The right hand side vector as well as the jacobian matrix.
        '''

        print('assembling the stiffness and jacobi matrix (scalar potential) (py)...')

        # the total number of nodes
        num_nodes = self.nodes.shape[0]

        # number of cell types
        #num_cell_types = len(self.cell_types)

        # make the rhs vector
        rhs = np.zeros((num_nodes, ))

        # make the sparse Jacobi matrix
        K = csr_array((num_nodes, num_nodes))
    
        # make the sparse Jacobi matrix
        dK = csr_array((num_nodes, num_nodes))

        # a global element counter (all materials)
        glob_el_cnt = 0

        # loop over all materials
        for n in range(self.num_materials):

            # get the geometry info of this volume
            cell_types = self.cell_types[n]
            cells = self.cells[n]

            # get also the permeability in this domain
            permeability = mu_list[n]

            # loop over cell types
            for i, ct in enumerate(cell_types):

                # make the finite element
                finite_element = FiniteElement(ct)

                # the numebr fo element nodes
                el_nodes = finite_element.get_number_of_nodes()

                # the number of elements of this type
                num_el = np.int32(len(cells[i]) / el_nodes)

                # get the quadrature nodes and weights
                q, w = self.get_quadrature_rule(ct, quad_order)

                # the number of quadrature points
                num_quad_pts = np.int32(len(q) / 3)

                # evaluate the shape functions at these points (not needed)
                # phi = finite_element.evaluate_basis(q)

                # evaluate the derivatives at these points
                d_phi = finite_element.evaluate_basis_derivative(q)

                # the number of triplets
                num_triplets = num_el*el_nodes*el_nodes

                # allocate the space for the triplet list (K)
                ij = np.zeros((num_triplets, 2), dtype=np.int32)
                vals_K = np.zeros((num_triplets, ))

                # allocate the space for the triplet list (J)
                vals_dK = np.zeros((num_triplets, ))

                # a triplet counter
                t_cnt = 0

                # loop over all elements
                for e_i in tqdm(range(num_el)):

                    # gmsh starts counting at 1
                    e = cells[i][e_i*el_nodes:(e_i+1)*el_nodes] - 1

                    # compute the Jacobian at the integration points
                    # inv_J is a numpy array of dimension (M x 3 x 3) where
                    # the inverse of the Jacobians are stored in the second and
                    # third dimensions
                    # M is the number of integration points
                    # det_J is an (M x 0) numpy array which stores the M values of det(J)
                    J = finite_element.compute_J(e, self.nodes, d_phi)
                    
                    inv_J = finite_element.compute_J_inv(J)
                    det_J = finite_element.compute_J_det(J)

                    # get the field vectors
                    this_H = H[glob_el_cnt*num_quad_pts:(glob_el_cnt+1)*num_quad_pts, :]

                    # compute also the field vector magnitude
                    H_mag = np.linalg.norm(this_H, axis=1)

                    # compute the permebility
                    mu = permeability.evaluate_mu(H_mag)

                    # compute the permebility derivative
                    d_mu = permeability.evaluate_mu_derivative(H_mag)

                    # loop over all combinations of basis functions
                    for j in range(el_nodes):
                        for k in range(j, el_nodes):

                            # this is the integration value (K)
                            int_val_rhs = 0.0

                            # this is the integration value (K)
                            int_val_K = 0.0

                            # this is the integration value (dK)
                            int_val_dK = 0.0       


                            # gaussian integration
                            for m in range(num_quad_pts):

                                # transform the gradients
                                grad_u = inv_J[m, :, :].T @ d_phi[m, j, :]
                                grad_v = inv_J[m, :, :].T @ d_phi[m, k, :]

                                # increment the integration value
                                int_val_rhs += mu[m]*det_J[m]*w[m]*np.sum(this_H[m, :]*grad_u)

                                # increment the integration value
                                int_val_K += mu[m]*det_J[m]*w[m]*np.sum(grad_u*grad_v)

                                # increment the integration value
                                if H_mag[m] >= 1e-14:
                                    
                                    int_val_dK += d_mu[m]*det_J[m]*w[m]*np.sum(grad_u*this_H[m, :])*np.sum(grad_v*this_H[m, :])/H_mag[m]
                                    
                            # sparse matrix version

                            # fill the triplet list
                            ij[t_cnt, 0] = e[j]
                            ij[t_cnt, 1] = e[k]
                            rhs[e[j]] += int_val_rhs
                            vals_K[t_cnt] = int_val_K
                            vals_dK[t_cnt] = int_val_dK
                            t_cnt += 1

                            # apply symmetry
                            if j != k:
                                ij[t_cnt, 0] = e[k]
                                ij[t_cnt, 1] = e[j]
                                rhs[e[k]] += int_val_rhs
                                vals_K[t_cnt] = int_val_K
                                vals_dK[t_cnt] = int_val_dK
                                t_cnt += 1

                    # increment the global element counter
                    glob_el_cnt += 1

                # print('number of triplets = {}'.format(len(vals)))

                # make the sparse matrix
                K += csr_array((vals_K, (ij[:, 0], ij[:, 1])), shape=(num_nodes, num_nodes))
                dK += csr_array((vals_dK, (ij[:, 0], ij[:, 1])), shape=(num_nodes, num_nodes))

            # add K and dK
            dK += K

        return rhs, dK    

    def compute_rhs(self, H, mu_list, quad_order=8):
        '''Compute the rhs and the jacobi matrix for the magnetic scalar potential for
        nonlinear problems. The H field vectors may include exterior field so that 
        this function can be used for the reduced vector potential formulation.
        
        This implementation is just to test the equations. We do not focus
        on performance at this point. A fast version of this code will be
        implemented in an optimized C++ code.

        :param x:
            The solution vector.
            
        :param mu_list:
            A list of permeability classes for each material.

        :param quad_order:
            The quadrature order. Default 8.

        :param H_ext:
            An external field for the reduced scalar potential formulation.
            If this array has zero rows (default), the external field is ignored.
            Otherwise, the number of rows must match the number of integration points.

        :return:
            The right hand side vector as well as the jacobian matrix.
        '''

        print('assembling the right hand side for the reduced scalar potential formulation (py)...')

        # the total number of nodes
        num_nodes = self.nodes.shape[0]

        # number of cell types
        #num_cell_types = len(self.cell_types)

        # make the rhs vector
        rhs = np.zeros((num_nodes, ))

        # make the sparse Jacobi matrix
        K = csr_array((num_nodes, num_nodes))
    
        # make the sparse Jacobi matrix
        dK = csr_array((num_nodes, num_nodes))

        # a global element counter (all materials)
        glob_el_cnt = 0

        # loop over all materials
        for n in range(self.num_materials):

            # get the geometry info of this volume
            cell_types = self.cell_types[n]
            cells = self.cells[n]

            # get also the permeability in this domain
            permeability = mu_list[n]

            # loop over cell types
            for i, ct in enumerate(cell_types):

                # make the finite element
                finite_element = FiniteElement(ct)

                # the numebr fo element nodes
                el_nodes = finite_element.get_number_of_nodes()

                # the number of elements of this type
                num_el = np.int32(len(cells[i]) / el_nodes)

                # get the quadrature nodes and weights
                q, w = self.get_quadrature_rule(ct, quad_order)

                # the number of quadrature points
                num_quad_pts = np.int32(len(q) / 3)

                # evaluate the shape functions at these points (not needed)
                # phi = finite_element.evaluate_basis(q)

                # evaluate the derivatives at these points
                d_phi = finite_element.evaluate_basis_derivative(q)

                # the number of triplets
                num_triplets = num_el*el_nodes*el_nodes

                # allocate the space for the triplet list (K)
                ij = np.zeros((num_triplets, 2), dtype=np.int32)
                vals_K = np.zeros((num_triplets, ))

                # allocate the space for the triplet list (J)
                vals_dK = np.zeros((num_triplets, ))

                # a triplet counter
                t_cnt = 0

                # loop over all elements
                for e_i in tqdm(range(num_el)):

                    # gmsh starts counting at 1
                    e = cells[i][e_i*el_nodes:(e_i+1)*el_nodes] - 1

                    # compute the Jacobian at the integration points
                    # inv_J is a numpy array of dimension (M x 3 x 3) where
                    # the inverse of the Jacobians are stored in the second and
                    # third dimensions
                    # M is the number of integration points
                    # det_J is an (M x 0) numpy array which stores the M values of det(J)
                    J = finite_element.compute_J(e, self.nodes, d_phi)
                    
                    inv_J = finite_element.compute_J_inv(J)
                    det_J = finite_element.compute_J_det(J)

                    # get the field vectors
                    this_H = H[glob_el_cnt*num_quad_pts:(glob_el_cnt+1)*num_quad_pts, :]

                    # compute also the field vector magnitude
                    H_mag = np.linalg.norm(this_H, axis=1)

                    # compute the permebility
                    mu = permeability.evaluate_mu(H_mag)

                    # loop over all combinations of basis functions
                    for j in range(el_nodes):

                        # this is the integration value (K)
                        int_val_rhs = 0.0

                        # gaussian integration
                        for m in range(num_quad_pts):

                            # transform the gradients
                            grad_u = inv_J[m, :, :].T @ d_phi[m, j, :]

                            # increment the integration value
                            int_val_rhs += mu[m]*det_J[m]*w[m]*np.sum(this_H[m, :]*grad_u)



                        rhs[e[j]] += int_val_rhs

                    # increment the global element counter
                    glob_el_cnt += 1

                # print('number of triplets = {}'.format(len(vals)))

        return rhs    

    def compute_stiffness_and_jacobi_matrix_c(self, x, mu_list, quad_order=8):
        '''Compute the stiffness and the jacobi matrix for the magnetic scalar potential for
        nonlinear problems.
        This implementation is the fast version of the above.
        
        :param x:
            The solution vector.
            
        :param mu_list:
            A list of permeability classes for each material.

        :param quad_order:
            The quadrature order. Default 8.

        :return:
            The stiffness matrix.
        '''

        # print('assembling the stiffness and jacobi matrix (scalar potential) (py)...')

        # the total number of nodes
        num_nodes = self.nodes.shape[0]

        # number of cell types
        #num_cell_types = len(self.cell_types)

        # make the sparse stiffness matrix
        K = csr_array((num_nodes, num_nodes))
    
        # make the sparse Jacobi matrix
        dK = csr_array((num_nodes, num_nodes))

        # loop over all materials
        for n in range(len(mu_list)):

            # get the geometry info of this volume
            cell_types = self.cell_types[n]
            cells = self.cells[n]

            # get also the permeability in this domain
            permeability = mu_list[n]

            # loop over cell types
            for i, ct in enumerate(cell_types):

                # make the finite element
                finite_element = FiniteElement(ct)

                # the numebr of element nodes
                el_nodes = finite_element.get_number_of_nodes()

                # the number of elements of this type
                num_el = np.int32(len(cells[i]) / el_nodes)

                # get the quadrature nodes and weights
                q, w = self.get_quadrature_rule(ct, quad_order)

                # the number of quadrature points
                num_quad_pts = np.int32(len(q) / 3)

                # evaluate the shape functions at these points (not needed)
                # phi = finite_element.evaluate_basis(q)

                # evaluate the derivatives at these points
                d_phi = finite_element.evaluate_basis_derivative(q)

                # call the cpp code (gmsh starts counting nodes at 1)
                my_K, my_dK = fem_c.compute_stiffness_and_jacobi_matrix(self.nodes, cells[i] - 1, d_phi, x, permeability, q, w)

                # increment the sparse matrix
                K += my_K
                dK += my_dK
                
            # add K and dK
            dK += K

        return K, dK
    
    def get_total_number_of_elements(self):
        '''Get the total number of elements in this FEM mesh.
        
        :return:
            The number of elements as integer.
            
        '''

        # the counter
        el_cnt = 0

        for n in range(self.num_materials):
            # get the geometry info of this volume
            cell_types = self.cell_types[n]
            cells = self.cells[n]
            # loop over cell types
            for i, ct in enumerate(cell_types):
                # make the finite element
                finite_element = FiniteElement(ct)

                # the numebr fo element nodes
                el_nodes = finite_element.get_number_of_nodes()

                # the number of elements of this type
                num_el = np.int32(len(cells[i]) / el_nodes)

                # increment the element counter
                el_cnt += num_el

        return el_cnt

    def get_total_number_of_evaluations(self, quad_order):
        '''Get the total number of evaluations in the FEM mesh for
        a certain quadrature order.

        :param quad_order:
            The quadrature order.
        
        :return:
            The number of elements as integer.
            
        '''

        # the counter
        eval_cnt = 0

        for n in range(self.num_materials):
            # get the geometry info of this volume
            cell_types = self.cell_types[n]
            cells = self.cells[n]
            # loop over cell types
            for i, ct in enumerate(cell_types):
                # make the finite element
                finite_element = FiniteElement(ct)

                # the numebr fo element nodes
                el_nodes = finite_element.get_number_of_nodes()

                # the number of elements of this type
                num_el = np.int32(len(cells[i]) / el_nodes)

                # get the quadrature nodes and weights
                _, w = self.get_quadrature_rule(ct, quad_order)

                # increment the element counter
                eval_cnt += num_el*len(w)

        return eval_cnt

    def compute_field_evaluation_matrix(self, quad_order=3):
        '''Compute the matrix to evaluate the field in the fem mesh.

        :param quad_order:
            The quadrature order. Default 8.

        :return:
            The sparse evaluation matrix.
        '''

        # the total number of nodes
        num_nodes = self.nodes.shape[0]

        # the total number of fem evaluations
        num_eval = self.get_total_number_of_evaluations(quad_order)

        # make space for the matrix
        X = csr_array((3*num_eval, num_nodes))

        # make space for the return points and field vectors
        points = np.zeros((0, 3))

        # loop over all materials
        for n in range(self.num_materials):
            
            # get the geometry info of this volume
            cell_types = self.cell_types[n]
            cells = self.cells[n]

            # loop over cell types
            for i, ct in enumerate(cell_types):

                # make the finite element
                finite_element = FiniteElement(ct)

                # the numebr fo element nodes
                el_nodes = finite_element.get_number_of_nodes()

                # the number of elements of this type
                num_el = np.int32(len(cells[i]) / el_nodes)

                # get the quadrature nodes and weights
                q, _ = self.matrix_factory.get_quadrature_rule(ct, quad_order)

                # the number of quadrature points
                num_quad_pts = np.int32(len(q) / 3)

                # evaluate the shape functions at these points
                phi = finite_element.evaluate_basis(q)

                # evaluate the derivatives at these points
                d_phi = finite_element.evaluate_basis_derivative(q)

                # make space for the evaluation points of this finite element type
                my_points = np.zeros((num_quad_pts*num_el, 3))

                # make space for the magnetic field of this finite element type
                my_H = np.zeros((num_quad_pts*num_el, 3))

                # call the cpp code (gmsh starts counting nodes at 1)

                # TO BE DONE...
                # my_X = fem_c.compute_field_evaluation_matrix(self.nodes, cells[i] - 1, d_phi, q)

                # increment matrix
                # X += my_X

        return X

    def compute_field(self, x, quad_order=3):
        '''Compute the field in the fem mesh.

        :param x:
            The solution vector.
    
        :param quad_order:
            The quadrature order. Default 8.

        :return:
            The evaluation points and the field vectors.
        '''
        
        # make space for the return points and field vectors
        points = np.zeros((0, 3))
        H = np.zeros((0, 3))
        B = np.zeros((0, 3))

        # get the mesh information
        nodes = self.nodes
    
        # loop over all materials
        for n in range(self.num_materials):

            
            # get the geometry info of this volume
            cell_types = self.cell_types[n]
            cells = self.cells[n]

            # loop over cell types
            for i, ct in enumerate(cell_types):

                # make the finite element
                finite_element = FiniteElement(ct)

                # the numebr fo element nodes
                el_nodes = finite_element.get_number_of_nodes()

                # the number of elements of this type
                num_el = np.int32(len(cells[i]) / el_nodes)

                # get the quadrature nodes and weights
                q, _ = self.get_quadrature_rule(ct, quad_order)

                # the number of quadrature points
                num_quad_pts = np.int32(len(q) / 3)

                # evaluate the shape functions at these points
                phi = finite_element.evaluate_basis(q)

                # evaluate the derivatives at these points
                d_phi = finite_element.evaluate_basis_derivative(q)

                # make space for the evaluation points of this finite element type
                my_points = np.zeros((num_quad_pts*num_el, 3))

                # make space for the magnetic field of this finite element type
                my_H = np.zeros((num_quad_pts*num_el, 3))

                # loop over all elements
                for e_i in range(num_el):

                    # gmsh starts counting at 1
                    e = cells[i][e_i*el_nodes:(e_i+1)*el_nodes] - 1

                    # evaluate this finite element for the global position
                    my_points[e_i*num_quad_pts:(e_i+1)*num_quad_pts, :] = finite_element.evaluate(e, nodes, phi)

                    # compute the Jacobian at the integration points
                    J = finite_element.compute_J(e, nodes, d_phi)

                    # we also need the inverse to transform
                    inv_J = finite_element.compute_J_inv(J)

                    # compute also the field vector
                    grad_phi, _ = self.compute_grad_phi(e, x, d_phi, inv_J)

                    for m in range(num_quad_pts):
                        # transform the gradient
                        my_H[e_i*num_quad_pts+m, :] = -1.0*grad_phi[m, :]

                # append to the return arrays
                points = np.append(points, my_points, axis=0)
                H = np.append(H, my_H, axis=0)

        return points, H
        

# class CurlCurlAssembler():

#     def __init__(self, mesh, volume_tags, material_list):
#         '''Default constructor.
        
#         :param mesh:
#             A gmsh mesh object.

#         :param volume_tags:
#             The gmsh tags for the volumes.
            
#         :param material_list:
#             A list of material properties.

#         :return:
#             None.
#         '''

#         # take the mesh
#         self.mesh = mesh

#         # set the element order
#         self.element_order = 1

#         # get some mesh info
#         _, _, num_edges, elementTags = get_mesh_info(self.mesh)

#         # the total number of dofs
#         self.num_dofs = num_edges

#         # get the materials list
#         self.material_list = material_list

#         # The nodes are not sorted correctly. I don't know why...
#         # But we need to get them like this:
#         node_tags, _, _ = mesh.getNodes()
#         num_nodes = len(node_tags)
#         node_tags = np.unique(node_tags)

#         # we now make an array of unique mesh nodes.
#         self.nodes = np.zeros((num_nodes, 3))

#         for i in range(num_nodes):
#             self.nodes[i, :] = mesh.getNode(node_tags[i])[0]
        
        
#         # the number of dofs per element. This is hardcoded at the momement.
#         # Only 1st order approximation is allowed
#         num_dof_el = get_num_edge_dofs(self.element_order)

#         # this is the function type string This is hardcoded at the momement.
#         # Only 1st order approximation is allowed
#         function_type = 'HcurlLegendre' + str(self.element_order-1)

#         # we allocate lists for the cell connectivities and types
#         # for all materials
#         self.num_materials = len(material_list)
#         self.cell_types = []
#         self.cells = []
#         self.global_ids = []
#         self.cell_tags = []

#         print('volume_tags = {}'.format(volume_tags))

#         if len(volume_tags) == 0:

#             # we take all materials
#             self.num_materials = 1

#             # get the elements
#             c_types_tmp, cell_tags, cells_tmp = gmsh.model.mesh.getElements(3, -1)

#             # append to the list
#             self.cell_types.append(c_types_tmp)
#             self.cells.append(cells_tmp)
#             self.cell_tags.append(cell_tags[0])

#             # the number of elements of this material
#             num_el = len(cell_tags[0])

#             # append to the global ids
#             self.global_ids.append(np.zeros((num_el, num_dof_el), dtype=np.int32))

#             # loop over the elements
#             for e in range(num_el):

#                 typeKeys, entityKeys, _ = mesh.getKeysForElement(cell_tags[0][e], function_type, returnCoord=False)
#                 self.global_ids[0][e, :] = entityKeys
        

#         else:
#             for i, tag in enumerate(volume_tags):
                
#                 # get the elements
#                 c_types_tmp, cell_tags, cells_tmp = gmsh.model.mesh.getElements(3, tag)

#                 # append to the list
#                 self.cell_types.append(c_types_tmp)
#                 self.cells.append(cells_tmp)
#                 self.cell_tags.append(cell_tags[0])
                
#                 # the number of elements of this material
#                 num_el = len(cell_tags[0])

#                 # append to the global ids
#                 self.global_ids.append(np.zeros((num_el, num_dof_el), dtype=np.int32))

#                 # loop over the elements
#                 for e in range(num_el):

#                     typeKeys, entityKeys, _ = mesh.getKeysForElement(cell_tags[0][e], function_type, returnCoord=False)
#                     self.global_ids[i][e, :] = entityKeys

#         # the number of nodes
#         num_nodes = self.nodes.shape[0]

#         return None
    
#     def setup_mesh(self, nodes, cells, cell_types):
#         """Setup the mesh from external information.

#         :param nodes:
#             The nodal coordinates.

#         :param cells:
#             The mesh connectivity list.

#         :param cell_types.
#             The mesh cell types list.

#         :return:
#             None.
#         """

#         self.nodes = nodes
#         self.cells = cells
#         self.cell_types = cell_types

#         return None
    

#     def compute_dF_curl_w(self, curl_w_hat, jacobians):

#         #number of quadrature points
#         num_points = np.int32(jacobians.shape[0]/9)

#         #number of basis functions
#         num_basis_fcns = np.int32(curl_w_hat.shape[0] / 3 / num_points)

#         #return values
#         dFcw = np.zeros((3,num_basis_fcns,num_points))

#         #temporal container for curl w
#         curl_w_tmp = np.zeros((3,num_basis_fcns))

#         for i in range(num_points):

#             curl_w_tmp[0,:] = curl_w_hat[0 + i*3*num_basis_fcns:(i+1)*3*num_basis_fcns:3]
#             curl_w_tmp[1,:] = curl_w_hat[1 + i*3*num_basis_fcns:(i+1)*3*num_basis_fcns:3]
#             curl_w_tmp[2,:] = curl_w_hat[2 + i*3*num_basis_fcns:(i+1)*3*num_basis_fcns:3]

#             dFcw[:,:,i] = np.array([[ jacobians[i*9]   , jacobians[i*9+3] , jacobians[i*9+6]],
#                                     [ jacobians[i*9+1] , jacobians[i*9+4] , jacobians[i*9+7]],
#                                     [ jacobians[i*9+2] , jacobians[i*9+5] , jacobians[i*9+8]]]) @ curl_w_tmp

#         return dFcw


#     def compute_stiffness_matrix(self, quad_order=8):
#         '''Compute the stiffness matrix using edge elements.
#         Can at the moment only handle meshes of a single element type.

#         :param quad_order:
#             The order of the quadrature rule.

#         :return:
#             The sparse stiffness matrix.
#         '''


#         print('compute stiffness matrix curl curl...')

#         # make the sparse stiffness matrix
#         K = csr_array((self.num_dofs, self.num_dofs))
    
#         # loop over all materials
#         for n in range(self.num_materials):
            
#             # get the geometry info of this volume
#             cell_types = self.cell_types[n]
#             cells = self.cells[n]

#             # get also the permeability in this domain
#             mat_prop = self.material_list[n]
            
#             # loop over cell types
#             for i, ct in enumerate(cell_types):

#                 # make the finite element
#                 finite_element = FiniteElement(ct)

#                 # the numebr fo element nodes
#                 el_nodes = finite_element.get_number_of_nodes()

#                 # the number of elements of this type
#                 num_el = np.int32(len(cells[i]) / el_nodes)

#                 # get the mesh connectivity
#                 c = cells[i].copy()
#                 c.shape = (num_el, el_nodes)

#                 # get the quadrature nodes and weights
#                 w, q = finite_element.get_quadrature_rule(quad_order)

#                 # the number of quadrature points
#                 num_quad_pts = np.int32(len(q) / 3)

#                 # evaluate the curls of the edge shape functions at the integration points
#                 curls, num_orientations = finite_element.get_curl_edge_basis_functions(self.mesh, q, ct, self.element_order)

#                 # evaluate the derivatives of the lagrange basis functions at these points
#                 d_phi = finite_element.evaluate_basis_derivative(q)

#                 # get the jacobians, determinants and orientations
#                 # _, _, _, orientations = get_vector_basis_mesh_info(self.mesh, q)
#                 jacobians, determinants, coordinates, orientations = get_vector_basis_mesh_info(self.mesh, q)

#                 # get the global ids
#                 glob_ids = self.global_ids[n]

#                 # the number of DoFs for the edge elements per finite element
#                 num_el_dofs = get_num_edge_dofs(self.element_order)

#                 # the number of triplets
#                 num_triplets = num_el*num_el_dofs*num_el_dofs

#                 # allocate the space for the triplet list
#                 ij = np.zeros((num_triplets, 2), dtype=np.int32)
#                 vals = np.zeros((num_triplets, ))

#                 # a triplet counter
#                 t_cnt = 0

#                 # loop over the finite elements
#                 for j, e in enumerate(c):

#                     # compute the Jacobian at the integration points
#                     # inv_J is a numpy array of dimension (M x 3 x 3) where
#                     # the inverse of the Jacobians are stored in the second and
#                     # third dimensions
#                     # M is the number of integration points
#                     # det_J is an (M x 0) numpy array which stores the M values of det(J)
#                     # J = finite_element.compute_J(e - 1, self.nodes, d_phi)
                    
#                     # inv_J = finite_element.compute_J_inv(J)
#                     # det_J = finite_element.compute_J_det(J)

#                     # this offset is the first index in the shape functions list for this orientation
#                     offset_shape_function = 3*num_quad_pts*num_el_dofs*orientations[0, j]

#                     # get the curls of the basis functions with this orientations
#                     curl_w_hat = curls[3*orientations[0, j]*num_quad_pts*num_el_dofs:3*(orientations[0, j]+1)*num_quad_pts*num_el_dofs]

#                     # print('Jac 1 = {}'.format(jacobians[j, :9]))
#                     # print('Jac 2 = {}'.format(J[0, :, :].T.flatten()))

#                     # compute the products dF.curl(w_hat)
#                     dFcw = self.compute_dF_curl_w(curl_w_hat, jacobians[j, :])
#                     # dFcw = self.compute_dF_curl_w(curl_w_hat, J.flatten())

#                     # to do:
#                     # compute nu

#                     #compute all products 1/det(dF)*(dF.w)^T v(B) (dF.w)
#                     for k in range(num_el_dofs):

#                         # curl_w_hat_k = np.array([])
#                         for l  in range(k, num_el_dofs):
                            
#                             # curl_w_hat_l = np.array([])

#                             # compute v.T A v
#                             vAv = np.sum(dFcw[:, k, :] * dFcw[:, l, :], axis = 0)
#                             this_prod = np.sum(vAv*w/determinants[j, :])
#                             # this_prod = np.sum(vAv*w/det_J)


#                             # fill the triplet list
#                             ij[t_cnt, 0] = glob_ids[j, k] - 1
#                             ij[t_cnt, 1] = glob_ids[j, l] - 1
#                             vals[t_cnt] = this_prod
#                             t_cnt += 1

#                             # apply symmetry
#                             if k != l:
#                                 ij[t_cnt, 0] = glob_ids[j, l] - 1
#                                 ij[t_cnt, 1] = glob_ids[j, k] - 1
#                                 vals[t_cnt] = this_prod
#                                 t_cnt += 1

#             K += csr_array((vals, (ij[:, 0], ij[:, 1])), shape=(self.num_dofs, self.num_dofs))

#         return K
    
#     def compute_stiffness_matrix_alt(self, quad_order=8):
#         '''Compute the stiffness matrix using edge elements.
#         Can at the moment only handle meshes of a single element type.

#         :param quad_order:
#             The order of the quadrature rule.

#         :return:
#             The sparse stiffness matrix.
#         '''


#         print('compute stiffness matrix curl curl...')

#         # all triplets
#         all_ij = np.zeros((0, 2), dtype=np.int64)
#         all_vals = np.zeros((0, ), dtype=float)

#         # make the sparse stiffness matrix
#         K = csr_array((self.num_dofs, self.num_dofs))
    
#         # loop over all materials
#         for n in range(self.num_materials):
            
#             # get the geometry info of this volume
#             cell_types = self.cell_types[n]
#             cells = self.cells[n]

#             # get the global orientations of the finite elements
#             orientations = get_vector_basis_orientations(self.mesh,
#                                                          self.cell_tags[n],
#                                                          element_spec='CurlHcurlLegendre' + str(self.element_order-1))
            
#             # get also the permeability in this domain
#             mat_prop = self.material_list[n]
            
#             # loop over cell types
#             for i, ct in enumerate(cell_types):

#                 # make the finite element
#                 finite_element = FiniteElement(ct)

#                 # the numebr fo element nodes
#                 el_nodes = finite_element.get_number_of_nodes()

#                 # the number of elements of this type
#                 num_el = np.int32(len(cells[i]) / el_nodes)

#                 # get the mesh connectivity
#                 c = cells[i].copy()
#                 c.shape = (num_el, el_nodes)

#                 # get the quadrature nodes and weights
#                 w, q = finite_element.get_quadrature_rule(quad_order)

#                 # the number of quadrature points
#                 num_quad_pts = np.int32(len(q) / 3)

#                 # evaluate the curls of the edge shape functions at the integration points
#                 curls, num_orientations = finite_element.get_curl_edge_basis_functions(self.mesh, q, ct, self.element_order)

#                 # evaluate the derivatives of the lagrange basis functions at these points
#                 d_phi = finite_element.evaluate_basis_derivative(q)

#                 # get the global ids
#                 glob_ids = self.global_ids[n]

#                 # the number of DoFs for the edge elements per finite element
#                 num_el_dofs = get_num_edge_dofs(self.element_order)

#                 # the number of triplets
#                 num_triplets = num_el*num_el_dofs*num_el_dofs

#                 # allocate the space for the triplet list
#                 ij = np.zeros((num_triplets, 2), dtype=np.int32)
#                 vals = np.zeros((num_triplets, ))

#                 # a triplet counter
#                 t_cnt = 0

#                 # loop over the finite elements
#                 for j, e in enumerate(c):


#                     # compute the Jacobian at the integration points
#                     # inv_J is a numpy array of dimension (M x 3 x 3) where
#                     # the inverse of the Jacobians are stored in the second and
#                     # third dimensions
#                     # M is the number of integration points
#                     # det_J is an (M x 0) numpy array which stores the M values of det(J)
#                     J = finite_element.compute_J(e - 1, self.nodes, d_phi)
                    
#                     # inv_J = finite_element.compute_J_inv(J)
#                     det_J = finite_element.compute_J_det(J)

#                     # this offset is the first index in the shape functions list for this orientation
#                     offset_shape_function = 3*num_quad_pts*num_el_dofs*orientations[j]

#                     # get the curls of the basis functions with this orientations
#                     curl_w_hat = curls[3*orientations[j]*num_quad_pts*num_el_dofs:3*(orientations[j]+1)*num_quad_pts*num_el_dofs]

#                     # to do:
#                     # compute nu

#                     # if 166 in glob_ids[j, :] :
#                     #     print('e = {}'.format(e))
#                     #     print('curl_w_hat = {}'.format(curl_w_hat))
#                     #     print('glob_ids[j, :] = {}'.format(glob_ids[j, :]))
#                     #     print('det_J = {}'.format(det_J))

#                     #compute all products 1/det(dF)*(dF.w)^T v(B) (dF.w)
#                     for k in range(num_el_dofs):

#                         for l  in range(k, num_el_dofs):
                            
#                             # the integration value
#                             int_val = 0.0

#                             for m in range(num_quad_pts):

                                
#                                 curl_w_hat_k = np.array([curl_w_hat[3*(m*num_el_dofs + k)     ],
#                                                          curl_w_hat[3*(m*num_el_dofs + k) + 1 ],
#                                                          curl_w_hat[3*(m*num_el_dofs + k) + 2 ]])

#                                 curl_w_hat_l = np.array([curl_w_hat[3*(m*num_el_dofs + l)     ],
#                                                          curl_w_hat[3*(m*num_el_dofs + l) + 1 ],
#                                                          curl_w_hat[3*(m*num_el_dofs + l) + 2 ]])
                                
#                                 dFcw_k = J[m, :, :] @ curl_w_hat_k
#                                 dFcw_l = J[m, :, :] @ curl_w_hat_l

#                                 int_val += np.sum(dFcw_k*dFcw_l)*w[m]/det_J[m]


#                             # fill the triplet list
#                             ij[t_cnt, 0] = glob_ids[j, k] - 1
#                             ij[t_cnt, 1] = glob_ids[j, l] - 1
#                             vals[t_cnt] = int_val
#                             t_cnt += 1

#                             # apply symmetry
#                             if k != l:
#                                 ij[t_cnt, 0] = glob_ids[j, l] - 1
#                                 ij[t_cnt, 1] = glob_ids[j, k] - 1
#                                 vals[t_cnt] = int_val
#                                 t_cnt += 1

#             all_ij = np.append(all_ij, ij, axis=0)
#             all_vals = np.append(all_vals, vals)

#         K = csr_array((all_vals, (all_ij[:, 0], all_ij[:, 1])), shape=(self.num_dofs, self.num_dofs))

#         return K

#     def compute_stiffness_and_jacobian_matrix(self, x, quad_order=8):
#         '''Compute the stiffness matrix using edge elements.
#         Can at the moment only handle meshes of a single element type.

#         :param x:
#             The solution vector at which the jacobian is calculated.

#         :param quad_order:
#             The order of the quadrature rule.

#         :return:
#             The sparse stiffness matrix.
#         '''


#         print('compute stiffness and jacobian matrix curl curl...')

#         # all triplets
#         all_ij = np.zeros((0, 2), dtype=np.int64)
#         all_vals = np.zeros((0, ), dtype=float)

#         # make the sparse stiffness matrix
#         K = csr_array((self.num_dofs, self.num_dofs))
    
#         # make the sparse jacobi matrix
#         dK = csr_array((self.num_dofs, self.num_dofs))

#         # loop over all materials
#         for n in range(self.num_materials):
            
#             # get the geometry info of this volume
#             cell_types = self.cell_types[n]
#             cells = self.cells[n]

#             # get the global orientations of the finite elements
#             orientations = get_vector_basis_orientations(self.mesh,
#                                                          self.cell_tags[n],
#                                                          element_spec='CurlHcurlLegendre' + str(self.element_order-1))
            
#             # get also the permeability in this domain
#             reluctance = self.material_list[n]
            
#             # loop over cell types
#             for i, ct in enumerate(cell_types):

#                 # make the finite element
#                 finite_element = FiniteElement(ct)

#                 # the numebr fo element nodes
#                 el_nodes = finite_element.get_number_of_nodes()

#                 # the number of elements of this type
#                 num_el = np.int32(len(cells[i]) / el_nodes)

#                 # get the mesh connectivity
#                 c = cells[i].copy()
#                 c.shape = (num_el, el_nodes)

#                 # get the quadrature nodes and weights
#                 w, q = finite_element.get_quadrature_rule(quad_order)

#                 # the number of quadrature points
#                 num_quad_pts = np.int32(len(q) / 3)

#                 # evaluate the curls of the edge shape functions at the integration points
#                 curls, num_orientations = finite_element.get_curl_edge_basis_functions(self.mesh, q, ct, self.element_order)

#                 # evaluate the derivatives of the lagrange basis functions at these points
#                 d_phi = finite_element.evaluate_basis_derivative(q)

#                 # get the global ids
#                 glob_ids = self.global_ids[n]

#                 # the number of DoFs for the edge elements per finite element
#                 num_el_dofs = get_num_edge_dofs(self.element_order)

#                 # the number of triplets
#                 num_triplets = num_el*num_el_dofs*num_el_dofs

#                 # allocate the space for the triplet list for K
#                 ij = np.zeros((num_triplets, 2), dtype=np.int32)
#                 vals_K = np.zeros((num_triplets, ))

#                 # allocate the space for the triplet list (J)
#                 vals_dK = np.zeros((num_triplets, ))

#                 # a triplet counter
#                 t_cnt = 0

#                 # loop over the finite elements
#                 for j, e in enumerate(c):


#                     # compute the Jacobian at the integration points
#                     # inv_J is a numpy array of dimension (M x 3 x 3) where
#                     # the inverse of the Jacobians are stored in the second and
#                     # third dimensions
#                     # M is the number of integration points
#                     # det_J is an (M x 0) numpy array which stores the M values of det(J)
#                     J = finite_element.compute_J(e - 1, self.nodes, d_phi)
                    
#                     # inv_J = finite_element.compute_J_inv(J)
#                     det_J = finite_element.compute_J_det(J)

#                     # this offset is the first index in the shape functions list for this orientation
#                     offset_shape_function = 3*num_quad_pts*num_el_dofs*orientations[j]

#                     # get the curls of the basis functions with this orientations
#                     curl_w_hat = curls[3*orientations[j]*num_quad_pts*num_el_dofs:3*(orientations[j]+1)*num_quad_pts*num_el_dofs]

#                     # compute the B field vector
#                     #B_vec, B_mag = self.compute_grad_phi(e, x, d_phi, inv_J)

#                     # compute the reluctance
#                     #nu = reluctance.evaluate_mu(B_mag)

#                     # compute the permebility derivative
#                     #d_mu = reluctance.evaluate_mu_derivative(B_mag)

#                     #compute all products 1/det(dF)*(dF.w)^T v(B) (dF.w)
#                     for k in range(num_el_dofs):

#                         for l  in range(k, num_el_dofs):
                            
#                             # the integration value
#                             int_val = 0.0

#                             for m in range(num_quad_pts):

                                
#                                 curl_w_hat_k = np.array([curl_w_hat[3*(m*num_el_dofs + k)     ],
#                                                          curl_w_hat[3*(m*num_el_dofs + k) + 1 ],
#                                                          curl_w_hat[3*(m*num_el_dofs + k) + 2 ]])

#                                 curl_w_hat_l = np.array([curl_w_hat[3*(m*num_el_dofs + l)     ],
#                                                          curl_w_hat[3*(m*num_el_dofs + l) + 1 ],
#                                                          curl_w_hat[3*(m*num_el_dofs + l) + 2 ]])
                                
#                                 dFcw_k = J[m, :, :] @ curl_w_hat_k
#                                 dFcw_l = J[m, :, :] @ curl_w_hat_l

#                                 int_val_K += nu[m]*np.sum(dFcw_k*dFcw_l)*w[m]/det_J[m]

#                                 # increment the integration value
#                                 if B_mag[m] >= 1e-14:
#                                     int_val_dK += d_nu[m]*np.sum(dFcw_k*B_vec[m, :])*np.sum(dFcw_l*B_vec[m, :])*w[m]/det_J[m]/B_mag[m]

#                             # fill the triplet list
#                             ij[t_cnt, 0] = glob_ids[j, k] - 1
#                             ij[t_cnt, 1] = glob_ids[j, l] - 1
#                             vals_K[t_cnt] = int_val_K
#                             vals_dK[t_cnt] = int_val_dK

#                             t_cnt += 1

#                             # apply symmetry
#                             if k != l:
#                                 ij[t_cnt, 0] = glob_ids[j, l] - 1
#                                 ij[t_cnt, 1] = glob_ids[j, k] - 1
#                                 vals_K[t_cnt] = int_val_K
#                                 vals_dK[t_cnt] = int_val_dK

#                                 t_cnt += 1

#             all_ij = np.append(all_ij, ij, axis=0)
#             all_vals_K = np.append(all_vals_K, vals_K)
#             all_vals_dK = np.append(all_vals_dK, vals_dK)

#         K = csr_array((all_vals_K, (all_ij[:, 0], all_ij[:, 1])), shape=(self.num_dofs, self.num_dofs))
#         dK = csr_array((all_vals_dK, (all_ij[:, 0], all_ij[:, 1])), shape=(self.num_dofs, self.num_dofs))

#         return K, dK + K

#     def compute_fcn_rhs(self, J_fcn, quad_order=8):
#         '''Compute the rhs for solving a magnetostatic problem using edge elements.
#         Can at the moment only handle meshes of a single element type.
#         This function is based on a current density function J_fcn.
        
#         :param J_fcn:
#             A function specifying the rhs.
        
#         :param quad_order:
#             The order of the quadrature rule.

#         :return:
#             The sparse stiffness matrix.
#         '''

#         print('compute rhs...')

#         # make the sparse stiffness matrix
#         rhs = np.zeros((self.num_dofs, ))

        
#         # loop over all materials
#         for n in range(self.num_materials):
            
#             # get the geometry info of this volume
#             cell_types = self.cell_types[n]
#             cells = self.cells[n]

#             # get the global orientations of the finite elements
#             orientations = get_vector_basis_orientations(self.mesh,
#                                                          self.cell_tags[n],
#                                                          element_spec='HcurlLegendre' + str(self.element_order-1))    

#             # get also the permeability in this domain
#             mat_prop = self.material_list[n]
            
#             # loop over cell types
#             for i, ct in enumerate(cell_types):

#                 # make the finite element
#                 finite_element = FiniteElement(ct)

#                 # the numebr fo element nodes
#                 el_nodes = finite_element.get_number_of_nodes()

#                 # the number of elements of this type
#                 num_el = np.int32(len(cells[i]) / el_nodes)

#                 # get the mesh connectivity
#                 c = cells[i].copy()
#                 c.shape = (num_el, el_nodes)

#                 # get the quadrature nodes and weights
#                 w, q = finite_element.get_quadrature_rule(quad_order)

#                 # the number of quadrature points
#                 num_quad_pts = np.int32(len(q) / 3)

#                 # evaluate the edge shape functions at the integration points
#                 basis, num_orientations = finite_element.get_edge_basis_functions(self.mesh, q, ct, self.element_order)

#                 # evaluate the shape functions at these points (not needed)
#                 phi = finite_element.evaluate_basis(q)

#                 # evaluate the derivatives of the lagrange basis functions at these points
#                 d_phi = finite_element.evaluate_basis_derivative(q)

#                 # get the global ids
#                 glob_ids = self.global_ids[n]

#                 # the number of DoFs for the edge elements per finite element
#                 num_el_dofs = get_num_edge_dofs(self.element_order)

                    
#                 # loop over the finite elements
#                 for j, e in enumerate(c):

#                     # compute the Jacobian at the integration points
#                     # inv_J is a numpy array of dimension (M x 3 x 3) where
#                     # the inverse of the Jacobians are stored in the second and
#                     # third dimensions
#                     # M is the number of integration points
#                     # det_J is an (M x 0) numpy array which stores the M values of det(J)
#                     J = finite_element.compute_J(e - 1, self.nodes, d_phi)
                
#                     inv_J = finite_element.compute_J_inv(J)
#                     det_J = finite_element.compute_J_det(J)

#                     # this offset is the first index in the shape functions list for this orientation
#                     offset_shape_function = 3*num_quad_pts*num_el_dofs*orientations[j]

#                     # get the curls of the basis functions with this orientations
#                     w_hat = basis[3*orientations[j]*num_quad_pts*num_el_dofs:3*(orientations[j]+1)*num_quad_pts*num_el_dofs]
                    

#                     # evaluate this finite element for the global position
#                     points = finite_element.evaluate(e - 1, self.nodes, phi)

#                     # evaluate current density
#                     j_eval = J_fcn(points)

#                     # compute all products 1/det(dF)*(dF.w)^T v(B) (dF.w)
#                     for k in range(num_el_dofs):

#                         # the integration value
#                         int_val = 0.0

#                         for m in range(num_quad_pts):

#                             # the basis function in local coordinates
#                             w_hat_k = np.array([w_hat[3*(m*num_el_dofs + k)     ],
#                                                      w_hat[3*(m*num_el_dofs + k) + 1 ],
#                                                      w_hat[3*(m*num_el_dofs + k) + 2 ]])

#                             # the basis function in global coordinates
#                             dFcw_k = inv_J[m, :, :].T @ w_hat_k

                            
#                             # the product J.w_tilde
#                             rhs[glob_ids[j, k] - 1] += np.sum(dFcw_k*j_eval[m, :])*w[m]*det_J[m]
    


#         return rhs

#     def compute_grad_phi(self, c, x, d_phi, inv_J):
#         '''Compute the magnetic field in a finite element
#         with connectivity c. The vector x stores the nodal 
#         values of the solution.
#         The array grad_phi stores the gradients of the FEM
#         basis functions.

#         :param c:
#             The node connectivity of the finite element.

#         :param x:
#             The solution vector.

#         :param d_phi:
#             The derivatives of all basis functions.

#         :param inv_J:
#             The inverse of the Jacobian matrix of the transformation.

#         :return:
#             The field vectors H, as well as the magnitude H_mag.
#         '''

#         # the number of evaluation points
#         num_eval = d_phi.shape[0]

#         # allocate the return array
#         grad_phi = np.zeros((num_eval, 3))

#         # evaluate
#         grad_phi[:, 0] = d_phi[:, :, 0] @ x[c]
#         grad_phi[:, 1] = d_phi[:, :, 1] @ x[c]
#         grad_phi[:, 2] = d_phi[:, :, 2] @ x[c]


#         # transform it to the global domain
#         for m in range(num_eval):
#             grad_phi[m, :] = inv_J[m, :, :].T @ grad_phi[m, :]


#         # return also the magnitude
#         mag = np.linalg.norm(grad_phi, axis=1)

#         return grad_phi, mag

#     def compute_B_in_element(self, glob_ids, x, curl_w_hat, J, det_J):
#         '''Compute the magnetic flux density in a finite element, given the solution vector
#         and the curls of the basis functions.

#         :param glob_ids:
#             The cell global edge basis function identifiers.

#         :param x:
#             The solution vector.

#         :param curl_w_hat:
#             The curl of the basis functions evaluated at the interpolation points.

#         :param J:
#             The jacobian matrices evaluated at the interpolation points.

#         :param det_J:
#             The determinants of the jacobians evaluated at the interpolation points.

#         :return:
#             The B field vectors at the interpolation points.
#         '''

#         # the number of interpolation points
#         num_points = len(det_J)

#         # the number of basis functions
#         num_el_dofs = len(curl_w_hat)/num_points/3

#         # make space for the return vector
#         B_ret = np.zeros((num_points, 3))

#         #compute all products 1/det(dF)*(dF.w)^T v(B) (dF.w)

#         for m in range(num_points):

#             for k in range(num_el_dofs):

#                 # the basis function in local coordinates
#                 curl_w_hat_k = np.array([curl_w_hat[3*(m*num_el_dofs + k)     ],
#                                          curl_w_hat[3*(m*num_el_dofs + k) + 1 ],
#                                          curl_w_hat[3*(m*num_el_dofs + k) + 2 ]])

#                 # the B field vector in global coordinates
#                 B_ret[m, :] += J[m, :, :] @ curl_w_hat_k * x[glob_ids[k] - 1] / det_J[m]

#         return B_ret

    
#     def compute_rhs_electric_potential(self, x_p, quad_order=8, select_mat='all'):
#         '''Compute the rhs for solving a magnetostatic problem using edge elements.
#         Can at the moment only handle meshes of a single element type.
#         This function is based on an electric potential discretized on the same mesh.
        
#         :param x_p:
#             The nodal values for the electric potential.
        
#         :param quad_order:
#             The order of the quadrature rule.

#         :param select_mat:
#             Select the material. Default 'all' means all materials.

#         :return:
#             The sparse stiffness matrix.
#         '''

#         print('compute rhs...')

#         # check if material selection is a string
#         if isinstance(select_mat, str):
#             if select_mat == 'all':
#                 materials = range(self.num_mat)
#         else:
#             materials = select_mat

#         # make the sparse stiffness matrix
#         rhs = np.zeros((self.num_dofs, ))
    
#         # loop over all materials
#         for n in materials:
            
#             # get the geometry info of this volume
#             cell_types = self.cell_types[n]
#             cells = self.cells[n]

#             # get also the permeability in this domain
#             mat_prop = self.material_list[n]
            
#             # get the global orientations of the finite elements
#             orientations = get_vector_basis_orientations(self.mesh,
#                                                          self.cell_tags[n],
#                                                          element_spec='HcurlLegendre' + str(self.element_order-1))    
            
#             # loop over cell types
#             for i, ct in enumerate(cell_types):

#                 # make the finite element
#                 finite_element = FiniteElement(ct)

#                 # the numebr fo element nodes
#                 el_nodes = finite_element.get_number_of_nodes()

#                 # the number of elements of this type
#                 num_el = np.int32(len(cells[i]) / el_nodes)

#                 # get the mesh connectivity
#                 c = cells[i].copy()
#                 c.shape = (num_el, el_nodes)

#                 # get the quadrature nodes and weights
#                 w, q = finite_element.get_quadrature_rule(quad_order)

#                 # the number of quadrature points
#                 num_quad_pts = np.int32(len(q) / 3)

#                 # evaluate the edge shape functions at the integration points
#                 basis, num_orientations = finite_element.get_edge_basis_functions(self.mesh, q, ct, self.element_order)

#                 # evaluate the shape functions at these points (not needed)
#                 phi = finite_element.evaluate_basis(q)

#                 # evaluate the derivatives of the lagrange basis functions at these points
#                 d_phi = finite_element.evaluate_basis_derivative(q)

#                 # get the global ids
#                 glob_ids = self.global_ids[n]

#                 # the number of DoFs for the edge elements per finite element
#                 num_el_dofs = get_num_edge_dofs(self.element_order)

#                 # loop over the finite elements
#                 for j, e in enumerate(c):


#                     # compute the Jacobian at the integration points
#                     # inv_J is a numpy array of dimension (M x 3 x 3) where
#                     # the inverse of the Jacobians are stored in the second and
#                     # third dimensions
#                     # M is the number of integration points
#                     # det_J is an (M x 0) numpy array which stores the M values of det(J)
#                     J = finite_element.compute_J(e - 1, self.nodes, d_phi)
                
#                     inv_J = finite_element.compute_J_inv(J)
#                     det_J = finite_element.compute_J_det(J)

#                     # this offset is the first index in the shape functions list for this orientation
#                     offset_shape_function = 3*num_quad_pts*num_el_dofs*orientations[j]

#                     # get the curls of the basis functions with this orientations
#                     w_hat = basis[3*orientations[j]*num_quad_pts*num_el_dofs:3*(orientations[j]+1)*num_quad_pts*num_el_dofs]
                    
#                     # evaluate this finite element for the global position
#                     points = finite_element.evaluate(e - 1, self.nodes, phi)

#                     # evaluate current density
#                     j_eval, _ = self.compute_grad_phi(e - 1, x_p, d_phi, inv_J)

#                     # compute all products 1/det(dF)*(dF.w)^T v(B) (dF.w)
#                     for k in range(num_el_dofs):

#                         # the integration value
#                         int_val = 0.0

#                         for m in range(num_quad_pts):

#                             # the basis function in local coordinates
#                             w_hat_k = np.array([w_hat[3*(m*num_el_dofs + k)     ],
#                                                      w_hat[3*(m*num_el_dofs + k) + 1 ],
#                                                      w_hat[3*(m*num_el_dofs + k) + 2 ]])

#                             # the basis function in global coordinates
#                             dFcw_k = inv_J[m, :, :].T @ w_hat_k

                            
#                             # the product J.w_tilde
#                             rhs[glob_ids[j, k] - 1] += np.sum(dFcw_k*j_eval[m, :])*w[m]*det_J[m]
    

#         return rhs


#     def get_quadrature_points(self, quad_order=3):
#         '''Get the quadrature points in the domain.

#         :param quad_order:
#             The order of the quadrature rule.

#         :return:
#             The points in an (M x 3) array.

#         '''

#         # make the return field evaluation points
#         points_ret = np.zeros((0, 3))

#         # marker for the domains
#         marker = np.zeros((0, ))

#         # loop over all materials
#         for n in range(self.num_materials):
            
#             # get the geometry info of this volume
#             cell_types = self.cell_types[n]
#             cells = self.cells[n]

#             # get also the permeability in this domain
#             mat_prop = self.material_list[n]
            
#             # loop over cell types
#             for i, ct in enumerate(cell_types):

#                 # make the finite element
#                 finite_element = FiniteElement(ct)

#                 # the numebr fo element nodes
#                 el_nodes = finite_element.get_number_of_nodes()

#                 # the number of elements of this type
#                 num_el = np.int32(len(cells[i]) / el_nodes)

#                 # get the mesh connectivity
#                 c = cells[i].copy()
#                 c.shape = (num_el, el_nodes)

#                 # get the quadrature nodes and weights
#                 w, q = finite_element.get_quadrature_rule(quad_order)

#                 # the number of quadrature points
#                 num_quad_pts = np.int32(len(q) / 3)

#                 # evaluate the shape functions at these points (not needed)
#                 phi = finite_element.evaluate_basis(q)

#                 # field evaluation points for this material
#                 this_points = np.zeros((num_el*num_quad_pts, 3))

#                 # loop over the finite elements
#                 for j, e in enumerate(c):

#                     # evaluate this finite element for the global position
#                     this_points[j*num_quad_pts:(j+1)*num_quad_pts, :] = finite_element.evaluate(e - 1, self.nodes, phi)

#                 points_ret = np.append(points_ret, this_points, axis=0)
#                 marker = np.append(marker, n*np.ones((this_points.shape[0], )), axis=0)
    
#         return points_ret, marker

#     def compute_B(self, x, quad_order=8):
#         '''Compute the B for a given solution vector.
        
#         :param x:
#             The solution vector.
        
#         :param quad_order:
#             The order of the quadrature rule.

#         :return:
#             The sparse stiffness matrix.
#         '''

#         print('compute B...')
#         print('Ho...')

#         # make the B field return vector
#         B_ret = np.zeros((0, 3))

#         # make the return field evaluation points
#         points_ret = np.zeros((0, 3))

#         # loop over all materials
#         for n in range(self.num_materials):
            
#             # get the geometry info of this volume
#             cell_types = self.cell_types[n]
#             cells = self.cells[n]

#             # get the global orientations of the finite elements
#             orientations = get_vector_basis_orientations(self.mesh,
#                                                          self.cell_tags[n],
#                                                          element_spec='CurlHcurlLegendre' + str(self.element_order-1))

#             # get also the permeability in this domain
#             mat_prop = self.material_list[n]
            
#             # loop over cell types
#             for i, ct in enumerate(cell_types):

#                 # make the finite element
#                 finite_element = FiniteElement(ct)

#                 # the numebr fo element nodes
#                 el_nodes = finite_element.get_number_of_nodes()

#                 # the number of elements of this type
#                 num_el = np.int32(len(cells[i]) / el_nodes)

#                 # get the mesh connectivity
#                 c = cells[i].copy()
#                 c.shape = (num_el, el_nodes)

#                 # get the quadrature nodes and weights
#                 w, q = finite_element.get_quadrature_rule(quad_order)

#                 # the number of quadrature points
#                 num_quad_pts = np.int32(len(q) / 3)

#                 # evaluate the curls of the edge shape functions at the integration points
#                 curls, num_orientations = finite_element.get_curl_edge_basis_functions(self.mesh, q, ct, self.element_order)

#                 # evaluate the shape functions at these points (not needed)
#                 phi = finite_element.evaluate_basis(q)

#                 # evaluate the derivatives of the lagrange basis functions at these points
#                 d_phi = finite_element.evaluate_basis_derivative(q)

#                 # get the global ids
#                 glob_ids = self.global_ids[n]

#                 # the number of DoFs for the edge elements per finite element
#                 num_el_dofs = get_num_edge_dofs(self.element_order)

#                 # B field vectors for this material
#                 this_B = np.zeros((num_el*num_quad_pts, 3))

#                 # field evaluation points for this material
#                 this_points = np.zeros((num_el*num_quad_pts, 3))

#                 # loop over the finite elements
#                 for j, e in enumerate(c):


#                     # compute the Jacobian at the integration points
#                     # inv_J is a numpy array of dimension (M x 3 x 3) where
#                     # the inverse of the Jacobians are stored in the second and
#                     # third dimensions
#                     # M is the number of integration points
#                     # det_J is an (M x 0) numpy array which stores the M values of det(J)
#                     J = finite_element.compute_J(e - 1, self.nodes, d_phi)
                    
#                     inv_J = finite_element.compute_J_inv(J)
#                     det_J = finite_element.compute_J_det(J)

#                     # this offset is the first index in the shape functions list for this orientation
#                     offset_shape_function = 3*num_quad_pts*num_el_dofs*orientations[j]

#                     # get the  basis functions with this orientations
#                     curl_w_hat = curls[3*orientations[j]*num_quad_pts*num_el_dofs:3*(orientations[j]+1)*num_quad_pts*num_el_dofs]

#                     # evaluate this finite element for the global position
#                     this_points[j*num_quad_pts:(j+1)*num_quad_pts, :] = finite_element.evaluate(e - 1, self.nodes, phi)

#                     #compute all products 1/det(dF)*(dF.w)^T v(B) (dF.w)
#                     for k in range(num_el_dofs):

#                         # the integration value
#                         int_val = 0.0

#                         for m in range(num_quad_pts):

#                             # the basis function in local coordinates
#                             curl_w_hat_k = np.array([curl_w_hat[3*(m*num_el_dofs + k)     ],
#                                                 curl_w_hat[3*(m*num_el_dofs + k) + 1 ],
#                                                 curl_w_hat[3*(m*num_el_dofs + k) + 2 ]])

#                             # the basis function in global coordinates
#                             this_B[j*num_quad_pts+m, :] += J[m, :, :] @ curl_w_hat_k * x[glob_ids[j, k] - 1] / det_J[m]


#                     B_test = self.compute_B_in_element(glob_ids[j, :], x, curl_w_hat, J, det_J)

#                     if j == 0:
#                         print('B_test = {}'.format(B_test))
#                         print('this_B = {}'.format(this_B[j*num_quad_pts:(j+1)*num_quad_pts, :]))

#                 B_ret = np.append(B_ret, this_B, axis=0)
#                 points_ret = np.append(points_ret, this_points, axis=0)
    
#         return points_ret, B_ret
    

#     def compute_A(self, x, quad_order=8):
#         '''Compute A for a given solution vector.
        
#         :param x:
#             The solution vector.
        
#         :param quad_order:
#             The order of the quadrature rule.

#         :return:
#             The sparse stiffness matrix.
#         '''

#         print('compute A...')

#         # make the A field return vector
#         A_ret = np.zeros((0, 3))

#         # make the return field evaluation points
#         points_ret = np.zeros((0, 3))

#         # loop over all materials
#         for n in range(self.num_materials):
            
#             # get the geometry info of this volume
#             cell_types = self.cell_types[n]
#             cells = self.cells[n]

#             # get the global orientations of the finite elements
#             orientations = get_vector_basis_orientations(self.mesh,
#                                                          self.cell_tags[n],
#                                                          element_spec='HcurlLegendre' + str(self.element_order-1))

#             # get also the permeability in this domain
#             mat_prop = self.material_list[n]
            
#             # loop over cell types
#             for i, ct in enumerate(cell_types):

#                 # make the finite element
#                 finite_element = FiniteElement(ct)

#                 # the numebr fo element nodes
#                 el_nodes = finite_element.get_number_of_nodes()

#                 # the number of elements of this type
#                 num_el = np.int32(len(cells[i]) / el_nodes)

#                 # get the mesh connectivity
#                 c = cells[i].copy()
#                 c.shape = (num_el, el_nodes)

#                 # get the quadrature nodes and weights
#                 w, q = finite_element.get_quadrature_rule(quad_order)

#                 # the number of quadrature points
#                 num_quad_pts = np.int32(len(q) / 3)

#                 # evaluate the curls of the edge shape functions at the integration points
#                 basis, num_orientations = finite_element.get_edge_basis_functions(self.mesh, q, ct, self.element_order)

#                 # evaluate the shape functions at these points (not needed)
#                 phi = finite_element.evaluate_basis(q)

#                 # evaluate the derivatives of the lagrange basis functions at these points
#                 d_phi = finite_element.evaluate_basis_derivative(q)

#                 # get the global ids
#                 glob_ids = self.global_ids[n]

#                 # the number of DoFs for the edge elements per finite element
#                 num_el_dofs = get_num_edge_dofs(self.element_order)

#                 # A field vectors for this material
#                 this_A = np.zeros((num_el*num_quad_pts, 3))

#                 # field evaluation points for this material
#                 this_points = np.zeros((num_el*num_quad_pts, 3))

#                 # loop over the finite elements
#                 for j, e in enumerate(c):


#                     # compute the Jacobian at the integration points
#                     # inv_J is a numpy array of dimension (M x 3 x 3) where
#                     # the inverse of the Jacobians are stored in the second and
#                     # third dimensions
#                     # M is the number of integration points
#                     # det_J is an (M x 0) numpy array which stores the M values of det(J)
#                     J = finite_element.compute_J(e - 1, self.nodes, d_phi)
                    
#                     inv_J = finite_element.compute_J_inv(J)
#                     det_J = finite_element.compute_J_det(J)

#                     # this offset is the first index in the shape functions list for this orientation
#                     offset_shape_function = 3*num_quad_pts*num_el_dofs*orientations[j]

#                     # get the curls of the basis functions with this orientations
#                     w_hat = basis[3*orientations[j]*num_quad_pts*num_el_dofs:3*(orientations[j]+1)*num_quad_pts*num_el_dofs]

#                     # evaluate this finite element for the global position
#                     this_points[j*num_quad_pts:(j+1)*num_quad_pts, :] = finite_element.evaluate(e - 1, self.nodes, phi)

#                     #compute all products 1/det(dF)*(dF.w)^T v(B) (dF.w)
#                     for k in range(num_el_dofs):

#                         # the integration value
#                         int_val = 0.0

#                         for m in range(num_quad_pts):

#                             # the basis function in local coordinates
#                             w_hat_k = np.array([w_hat[3*(m*num_el_dofs + k)     ],
#                                                 w_hat[3*(m*num_el_dofs + k) + 1 ],
#                                                 w_hat[3*(m*num_el_dofs + k) + 2 ]])

#                             # the basis function in global coordinates
#                             this_A[j*num_quad_pts+m, :] += inv_J[m, :, :].T @ w_hat_k * x[glob_ids[j, k] - 1]

#                 A_ret = np.append(A_ret, this_A, axis=0)
#                 points_ret = np.append(points_ret, this_points, axis=0)
    
#         return points_ret, A_ret


import numpy as np
import gmsh
import pyvista as pv
import matplotlib.pyplot as plt
import networkx as nx

def get_mesh_info(mesh, element_order=1):
    '''Get some information about the mesh.
    
    :param mesh:
        A gmsh mesh object.

    :param element_order:
        The order of the finite element.

    :return:
        The number of degrees of freedom (assuming a certain
        fem approximation space), the number of faces,
        the number of edges and the element tags.
    '''
    
    # get element information
    elementTypes, elementTags, elementNodes = mesh.getElements(dim=3)
    elementTypes = elementTypes[0]
    elementTags = elementTags[0]
    elementNodes = elementNodes[0]

    # number of elements
    num_elements = elementTags.shape[0]

    # get edge information
    mesh.createEdges()
    edgeTags, edgeNodes = mesh.getAllEdges()

    # number of edges
    num_edges = len(edgeTags)

    # get face information
    mesh.createFaces()

    # get faces
    faceTags, faceNodes = mesh.getAllFaces(3)

    # number of faces
    num_faces = len(faceTags)

    # process information
    
    if element_order == 1:
        num_dofs = num_edges
    elif element_order == 2:
        num_dofs = 2*num_edges
    elif element_order == 3:
        num_dofs = 3*num_edges + 4*num_faces + 6*num_elements
    elif element_order == 4:
        num_dofs = 3*num_edges + 4*num_faces + 6*num_elements
    else:
        print('Warning! Element order {} not implemented!')
        num_dofs = -1

    return num_dofs, num_faces, num_edges, elementTags

def get_vector_basis_mesh_info(mesh, loc):
    '''Get all mesh information needed for the matrix assembly using
    vectorial finite elements, for instance Hcurl conforming ones.

    :param mesh:
        A gmsh mesh object.

    :param int_order:
        The integration local coordinates.

    :return:
        The jacobians at the integration points,
        the determinants at the integration points,
        the physical coordinates at the integration points,
        the orientation of the finite elements.
    '''

    # get some mesh info
    _, _, _, elementTags = get_mesh_info(mesh)
    num_el = elementTags.shape[0]

    # number of integration points
    num_points = np.int64(len(loc)/3)
    # jacobians data container
    jacobians = np.zeros((num_el, num_points*9))
    # determinants container
    determinants = np.zeros((num_el, num_points))
    # coordinates container
    coordinates = np.zeros((num_el, num_points*3))
    # orientations container
    orientations = np.zeros((1, num_el), dtype=np.int32)

    # loop over the elements
    for e in range(num_el):

        # evaluate the finite elements
        jacobians[e, :], determinants[e, :], coordinates[e, :] = mesh.getJacobian(elementTags[e], loc)

        # get the orientation of this element
        orientations[0, e] = mesh.getBasisFunctionsOrientationForElement(elementTags[e], "CurlHcurlLegendre2")

    return jacobians, determinants, coordinates, orientations

def get_vector_basis_orientations(mesh, element_tags, element_spec='HcurlLegendre1'):
    '''Get all element orientations needed for the matrix assembly using
    vectorial finite elements, for instance Hcurl conforming ones.

    :param mesh:
        A gmsh mesh object.

    :param element_tags:
        The gmsh element tag.

    :param element_spec:
        The finite element gmsh specifyer.

    :return:
        the orientation of the finite elements.
    '''

    # get some mesh info
    num_el = len(element_tags)

    # orientations container
    orientations = np.zeros((num_el, ), dtype=np.int32)

    # loop over the elements
    for i in range(num_el):

        # get the orientation of this element
        orientations[i] = mesh.getBasisFunctionsOrientationForElement(element_tags[i], element_spec)

    return orientations

def get_num_edge_dofs(order, element_type='Tet'):
    '''Get the number of degrees of freedom per element for
    a certain finite element order.

    :param element_order:
        The order of the finite element.

    :param element_type:
        The finite element type. Currently only Tet is supported!

    :return:
        The number of dofs.
    '''

    if element_type == 'Tet':
        if order == 1:
            return 6
        elif order == 2:
            return 12
        elif order == 3:
            return 30
        else:
            print('Tet edge elements of order {} unsupported!'.format(order))
            return -1

    elif element_type == 'Hex':
        if order == 1:
            return 12
        elif order == 2:
            return 54
        elif order == 3:
            return 144
        else:
            print('Hex edge elements of order {} unknown!'.format(order))
            return -1
        
    else:
        print('Edge elements of type {} unknown!'.format(element_type))


def get_global_ids_for_entities(type_keys, entity_keys, num_el, num_edges, num_faces, order):
    '''Function to compute the global identification numbers of degrees of freedom.

    :param type_keys:
        The type keys.

    :param entity_keys:
        The entity keys.

    :param num_edges:
        The number of edges in the mesh.

    :param num_faces:
        The number of faces in the mesh.

    :param order:
        The order of the edge elements.

    :return:
        The global ids array.

    '''


    if order == 1:
        # all type keys are 1. i.e. edge functions
        return entity_keys

    elif order == 2:
        # type keys are:
        # 1: edge function (1st kind)
        # 2: edge function (2nd kind)

        return np.int32(entity_keys + (type_keys == 2)*num_edges)
        
    elif order == 3:
        # type keys are:
        # 1: edge function (1st kind)
        # 2: edge function (2nd kind)

        return np.int32(entity_keys + (type_keys == 2)*num_edges \
                + (type_keys == 3)*2*num_edges \
                + (type_keys == 4)*(2*num_edges + num_faces) \
                + (type_keys == 5)*(2*num_edges  + 2*num_faces)\
                + (type_keys == 6)*(2*num_edges  + 3*num_faces)\
                + (type_keys == 7)*(2*num_edges  + 4*num_faces)\
                + (type_keys == 8)*(2*num_edges  + 4*num_faces + num_el)\
                + (type_keys == 9)*(2*num_edges  + 4*num_faces + 2*num_el)\
                + (type_keys == 10)*(2*num_edges  + 4*num_faces + 3*num_el)\
                + (type_keys == 11)*(2*num_edges  + 4*num_faces + 4*num_el)\
                + (type_keys == 12)*(2*num_edges  + 4*num_faces + 5*num_el))

    elif order == 4:
        # type keys are:
        # 0: vertex function

        # 1: edge function (1st kind)
        # 2: edge function (2nd kind)

        # 3: face function (1st kind)
        # 4: face function (2nd kind)
        # 5: face function (3rd kind)
        # 6: face function (4th kind)

        # 7: bubble function (1st kind)
        # 8: bubble function (2nd kind)
        # 9: bubble function (3rd kind)
        # 10: bubble function (4th kind)
        # 11: bubble function (5th kind)
        # 12: bubble function (6th kind)

        return np.int32(entity_keys + (type_keys == 2)*num_edges \
                + (type_keys == 3)*2*num_edges \
                + (type_keys == 4)*(2*num_edges + num_faces) \
                + (type_keys == 5)*(2*num_edges  + 2*num_faces)\
                + (type_keys == 6)*(2*num_edges  + 3*num_faces)\
                + (type_keys == 7)*(2*num_edges  + 4*num_faces)\
                + (type_keys == 8)*(2*num_edges  + 4*num_faces + num_el)\
                + (type_keys == 9)*(2*num_edges  + 4*num_faces + 2*num_el)\
                + (type_keys == 10)*(2*num_edges  + 4*num_faces + 3*num_el)\
                + (type_keys == 11)*(2*num_edges  + 4*num_faces + 4*num_el)\
                + (type_keys == 12)*(2*num_edges  + 4*num_faces + 5*num_el))
        
    elif order == 5:
        # type keys are:
        # 0: vertex function

        # 1: edge function (1st kind)
        # 2: edge function (2nd kind)
        # 3: edge function (3rd kind)

        # 4: face function (1st kind)
        # 5: face function (2nd kind)
        # 6: face function (3rd kind)
        # 7: face function (4th kind)
        # 8: face function (5th kind)
        # 9: face function (6th kind)
        # 10: face function (7th kind)
        # 11: face function (8th kind)
        # 12: face function (9th kind)
        # 13: face function (10th kind)
        # 14: face function (11th kind)
        # 15: face function (12th kind)

        # 16: bubble function (1st kind)
        # 17: bubble function (2nd kind)
        # 18: bubble function (3rd kind)
        # 19: bubble function (4th kind)
        # 20: bubble function (5th kind)
        # ...
        # 
        # 51: bubble function (36th kind)
            
        ret_keys = entity_keys.copy()

        offset = num_edges
        # edge basis functions
        for type in range(2, 4):
            ret_keys[type_keys == type] += offset
            offset += num_edges

        # face basis functions
        for type in range(4, 16):
            ret_keys[type_keys == type] += offset
            offset += num_faces

        # bubble functions
        for type in range(16, 52):
            ret_keys[type_keys == type] += offset
            offset += num_el
            
        return ret_keys
        
    else:
        print('Edge elements of order {} are not implemented yet!'.format(order))
        return -1

def get_global_ids(mesh, element_order):
    '''Get the global ids for the edge basis functions of a certain order.

    :param mesh:
        A gmsh mesh object.

    :param element_order:
        The order of the finite element.

    :return:
        The global ids.
    '''
        
    # get some mesh info
    _, num_faces, num_edges, elementTags = get_mesh_info(mesh)
    num_el = elementTags.shape[0]

    # the number of dofs per element
    num_dof_el = get_num_edge_dofs(element_order)

    # global ids container
    global_ids = np.zeros((num_el, num_dof_el), dtype=np.int32)

    # this is the function type string
    function_type = 'HcurlLegendre' + str(element_order-1)

    # loop over the elements
    for e in range(num_el):

        #get the global ids
        typeKeys, entityKeys, _ = mesh.getKeysForElement(elementTags[e], function_type, returnCoord=False)

        global_ids[e, :] = get_global_ids_for_entities(typeKeys, entityKeys, num_el, num_edges, num_faces, element_order)

    return global_ids

def get_edge_boundary_dofs(model, boundary_names, element_order=1):
    '''Get edge elements boundary DoFs.

    :param model:
        A gmsh model object.

    :param boundary_names:
        The name tags of the Dirichlet boundaries.

    :return:
        The indices of the edge_dofs and face_dofs
    '''
    # create all edges and faces
    model.mesh.createEdges()
    model.mesh.createFaces()

    # get all physical groups corresponding to surfaces
    boundary_groups = model.getPhysicalGroups(2)

    # FOR THE FUTURE
    # check if they are Dirichlet or Neumann type
    # so far only Dirichlet is implemented

    # number of boundaries
    num_boundaries = len(boundary_groups)

    num_dofs, num_faces,num_edges, elementTags = get_mesh_info(model.mesh)
    num_el = elementTags.shape[0]

    # containers for edge and face DoFs
    boundary_dofs = []
    # edge_dofs = []
    # face_dofs = []

    #function_space_string = 'HcurlLegendre{}'.format(element_order-1)
    function_space_string = 'HcurlLegendre{}'.format(element_order)

    # function space string
    if element_order > 3:
        print('Element order {} not supported!'.format(element_order))
        
    for bn in boundary_names:

        # get the dim tags
        dim_tags = model.getEntitiesForPhysicalName(bn)

        # loop over the boundaries
        for i, dt in enumerate(dim_tags):

            # boundary faces info
            b_face_types, b_face_tags, b_face_nodes = model.mesh.getElements(dt[0], dt[1])

            # loop over the faces
            for j in range(len(b_face_tags[0])):

                # get the keys for the DoFs of this face
                b_face_type_keys, b_face_entity_keys, b_face_coord = model.mesh.getKeysForElement(b_face_tags[0][j], function_space_string)

                glob_ids = get_global_ids_for_entities(b_face_type_keys, b_face_entity_keys, num_el, num_edges, num_faces, element_order)

                # make global indices based on type and face key
                for k in range(len(b_face_type_keys)):

                    # construct the global id
                    # global_id = np.int32(b_face_entity_keys[k])
                    global_id = np.int32(glob_ids[k])

                    boundary_dofs.append(global_id)
                    # if b_face_type_keys[k] < 4:
                    #     edge_dofs.append(global_id)
                    # else:
                    #    face_dofs.append(global_id)

    # make them unique
    boundary_dofs = np.unique(np.array(boundary_dofs))
    # face_dofs = np.unique(np.array(face_dofs))

    return boundary_dofs


def setup_coil_terminals(model, domain_name, pos_terminals):
    '''Setup the terminals for a coil.

    :param model:
        The gmsh model.

    :param domain_name:
        The name of the coil domain.

    :param pos_terminals:
        The center positions of the two terminals.

    :return:
        None.
    '''

    # get the two terminals
    vol = model.getEntitiesForPhysicalName(domain_name)[0]
    
    # get the boundary of the coil domain
    boundary = model.getBoundary([vol])


    # loop over the boundary surfaces
    for i, bs in boundary:
        
        normal = model.getNormal(abs(bs), [0., 0.])

        diff = np.linalg.norm(np.array([1., 0., 0.]) - np.array([abs(normal[0]), normal[1], normal[2]]))

        if diff < 1e-12:
            if model.isInside(2, abs(bs), pos_terminals[0]):
                model.addPhysicalGroup(2, [abs(bs)], name=domain_name + '_terminal_1')
            elif model.isInside(2, abs(bs), pos_terminals[1]):
                model.addPhysicalGroup(2, [abs(bs)], name=domain_name + '_terminal_2')
    
    return 

def get_cotree_dofs(gmsh_mesh):
    '''Get the degrees of freedom of the co-treee.

    :param gmsh_mesh:
        The gmsh mesh object.

    :return:
        The degrees of freedom of the cotree.
    '''

    # make all edges
    gmsh_mesh.createEdges()

    # get the edge connectivity
    edge_tags, edge_nodes  = gmsh_mesh.get_all_edges()
    edge_nodes.shape = (len(edge_tags), 2)
    
    nodes = np.zeros((len(edge_tags), 3))

    # setup a graph object
    G = nx.Graph()
    
    # this map is needed to get back to the node tags
    edge_id_map = {}
    # add all edges
    for i, e in enumerate(edge_nodes):
        edge_tuple = tuple(sorted((e[0], e[1])))
        G.add_edge(edge_tuple[0], edge_tuple[1], weight=1)
        edge_id_map[edge_tuple] = edge_tags[i]
    
    # Ensure the graph is connected
    if not nx.is_connected(G):
        raise ValueError("The input graph must be connected.")

    # get the spanning tree
    spanning_tree = nx.maximum_spanning_tree(G,algorithm='boruvka')
    spanning_tree_edges = list(spanning_tree.edges())

    # get the co-tree
    all_edges = set(G.edges())
    tree_edges = set(spanning_tree_edges)
    cotree_edges = list(all_edges - tree_edges)


    print('number of tree edges = {}'.format(len(spanning_tree_edges)))
    print('number of co-tree edges = {}'.format(len(cotree_edges)))


    tree_edge_dofs = []
    for edge in list(tree_edges):
        tree_edge_dofs.append(edge_id_map[tuple(sorted((edge[0], edge[1])))])

    co_tree_edge_dofs = []
    for edge in list(all_edges - tree_edges):
        co_tree_edge_dofs.append(edge_id_map[tuple(sorted((edge[0], edge[1])))])

    tree_edge_dofs = np.array(tree_edge_dofs, dtype=np.int64) - 1
    co_tree_edge_dofs = np.array(co_tree_edge_dofs, dtype=np.int64) - 1

    return tree_edge_dofs, co_tree_edge_dofs


import numpy as np
import gmsh

from .mesh_tools import get_vector_basis_orientations
from .finite_element import FiniteElement

def compute_grad_phi(c, x, d_phi, inv_J):
    '''Compute the magnetic field in a finite element
    with connectivity c. The vector x stores the nodal 
    values of the solution.
    The array grad_phi stores the gradients of the FEM
    basis functions.

    :param c:
        The node connectivity of the finite element.

    :param x:
        The solution vector.

    :param d_phi:
        The derivatives of all basis functions.

    :param inv_J:
        The inverse of the Jacobian matrix of the transformation.

    :return:
        The field vectors H, as well as the magnitude H_mag.
    '''

    # the number of evaluation points
    num_eval = d_phi.shape[0]

    # allocate the return array
    grad_phi = np.zeros((num_eval, 3))

    # evaluate
    grad_phi[:, 0] = d_phi[:, :, 0] @ x[c]
    grad_phi[:, 1] = d_phi[:, :, 1] @ x[c]
    grad_phi[:, 2] = d_phi[:, :, 2] @ x[c]


    # transform it to the global domain
    for m in range(num_eval):
        grad_phi[m, :] = inv_J[m, :, :].T @ grad_phi[m, :]


    # return also the magnitude
    mag = np.linalg.norm(grad_phi, axis=1)

    return grad_phi, mag

def evaluate_laplace_solution(positions, gmsh_model, vol_tags, solution):
    '''Compute the B field (interior or exterior) at certain posisions.

    :param positions:
        A numpy array of size (M x 3) with the coordinates in the columns.

    :param gmsh_model:
        The gmsh model object.

    :param vol_tags:
        The gmsh volume tags.
            
    :param solution:
        A solution vector.

    :return:
        A numpy matrix of dimension (M x 3) with the B field components in the
        columns. 
    '''

    # the number of positions
    num_pos = positions.shape[0]

    # The nodes are not sorted correctly. I don't know why...
    # But we need to get them like this:
    node_tags, _, _ = gmsh_model.mesh.getNodes()
    num_nodes = len(node_tags)
    node_tags = np.unique(node_tags)

    # we now make an array of unique mesh nodes.
    nodes = np.zeros((num_nodes, 3))

    for i in range(num_nodes):
        nodes[i, :] = gmsh_model.mesh.getNode(node_tags[i])[0]

    # we now allocate a return vector
    field = np.zeros((num_pos, 3))

    # domain specifier for the volume number
    domain_spec = np.zeros((num_pos, ), dtype=np.int32)

    # this is a mask for the outside points
    outside_mask = np.zeros((num_pos, ), dtype=bool)

    # for some reason, gmsh renumbers nodes...
    # we need to make a mapping between gmsh node tags and my node tags
    node_map = np.zeros((num_nodes, ), dtype=np.int64)

    # loop over all nodes
    for i in range(num_nodes):
        # get the coordinates
        coord, _, _, _ = gmsh_model.mesh.get_node(i+1)
        # search for this nodes in my nodes
        diff_x = coord[0] - nodes[:, 0]
        diff_y = coord[1] - nodes[:, 1]
        diff_z = coord[2] - nodes[:, 2]
        # the distance
        dist = np.sqrt(diff_x*diff_x + diff_y*diff_y + diff_z*diff_z)
        # get the minimum
        node_map[i] = np.argmin(dist)


    # we loop over all of the positions
    for i in range(num_pos):

        for j, v in enumerate(vol_tags):
            
            # check if this point is inside or outside
            is_inside = gmsh_model.is_inside(3, v, positions[i, :])

            if is_inside == True:
                domain_spec[i] = j + 1
                break

        if is_inside == True:
            # print('position {}'.format(positions[i, :]))
            # get the element and local coordinates
            el, type, node_tags, u, v, w = gmsh_model.mesh.get_element_by_coordinates(positions[i, 0], positions[i, 1], positions[i, 2], dim=3)

            # the node indices of my numbering
            node_indx = node_map[node_tags-1]

            # make the finite element
            finite_element = FiniteElement(type)

            # evaluate the derivatives at these points
            d_phi = finite_element.evaluate_basis_derivative(np.array([[u, v, w]]).flatten())
            # d_phi = d_phi[:, indx, :]
            # compute the Jabcobi matrix
            J = finite_element.compute_J(node_indx, nodes, d_phi)
                
            # invert the Jacobian
            inv_J = finite_element.compute_J_inv(J)

            # compute also the field vector
            field[i, :], _ = compute_grad_phi(node_indx, solution, d_phi, inv_J)

        else:

            # mark this point as outside
            outside_mask[i] = True

    return field

def evaluate_curl_curl_solution(positions, gmsh_model, vol_tags, solution, glob_ids):
    '''Evaluate the B field (interior or exterior) at certain posisions.
    This function is to use for the vector potential formulation curl curl equation.

    :param positions:
         A numpy array of size (M x 3) with the coordinates in the columns.

    :param gmsh_model:
        The gmsh model object.

    :param vol_tags:
        The gmsh volume tags.
                
    :param solution:
        A solution vector.

    :return:
        A numpy matrix of dimension (M x 3) with the B field components in the
        columns. 
    '''


    # the number of positions
    num_pos = positions.shape[0]

    print('evaluate field at {} positions...'.format(num_pos))

    # The nodes are not sorted correctly. I don't know why...
    # But we need to get them like this:
    node_tags, _, _ = gmsh_model.mesh.getNodes()
    num_nodes = len(node_tags)
    node_tags = np.unique(node_tags)

    # we now make an array of unique mesh nodes.
    nodes = np.zeros((num_nodes, 3))

    for i in range(num_nodes):
        nodes[i, :] = gmsh_model.mesh.getNode(node_tags[i])[0]

    # we now allocate a return vector
    B_ret = np.zeros((num_pos, 3))

    # domain specifier for the volume number
    domain_spec = np.zeros((num_pos, ), dtype=np.int32)

    # this is a mask for the outside points
    outside_mask = np.zeros((num_pos, ), dtype=bool)

    # for some reason, gmsh renumbers nodes...
    # we need to make a mapping between gmsh node tags and my node tags
    node_map = np.zeros((num_nodes, ), dtype=np.int64)

    # loop over all nodes
    for i in range(num_nodes):
        # get the coordinates
        coord, _, _, _ = gmsh_model.mesh.get_node(i+1)
        # search for this nodes in my nodes
        diff_x = coord[0] - nodes[:, 0]
        diff_y = coord[1] - nodes[:, 1]
        diff_z = coord[2] - nodes[:, 2]
        # the distance
        dist = np.sqrt(diff_x*diff_x + diff_y*diff_y + diff_z*diff_z)
        # get the minimum
        node_map[i] = np.argmin(dist)


    # the number of edge basis functions per element
    num_el_dofs = 6
    

    # we loop over all of the positions
    for i in range(num_pos):

        for j, v in enumerate(vol_tags):
                
            # check if this point is inside or outside
            is_inside = gmsh_model.is_inside(3, v, positions[i, :])

            if is_inside == True:
                domain_spec[i] = j + 1
                break

            # print('is inside = {}'.format(is_inside))

        if is_inside == True:

            # print('position {}'.format(positions[i, :]))

            # get the element and local coordinates
            el, type, node_tags, u, v, w = gmsh_model.mesh.get_element_by_coordinates(positions[i, 0], positions[i, 1], positions[i, 2], dim=3)


            # print('element = {}'.format(el))            

            # the index of this element in the orientations table
            el_index = el - 1

            # the node indices of my numbering
            node_indx = node_map[node_tags-1]

            # make the finite element
            finite_element = FiniteElement(type)

            # evaluate the derivatives at these points
            d_phi = finite_element.evaluate_basis_derivative(np.array([[u, v, w]]).flatten())
            # d_phi = d_phi[:, indx, :]
            # compute the Jabcobi matrix
            J = finite_element.compute_J(node_indx, nodes, d_phi)
                    
            # invert the Jacobian
            det_J = finite_element.compute_J_det(J)

            # evaluate the curls of the edge shape functions at the evaluation points
            curls, num_orientations = finite_element.get_curl_edge_basis_functions(gmsh_model.mesh, np.array([[u, v, w]]).flatten(), type, 1)
            
            # get the global orientations of the finite elements
            orientation = gmsh_model.mesh.getBasisFunctionsOrientationForElement(el, 'CurlHcurlLegendre0') 


            # print('orientation = {}'.format(orientation))

            # get the  basis functions with this orientations
            curl_w_hat = curls[3*orientation*num_el_dofs:3*(orientation+1)*num_el_dofs]
            
            # get the keys for this entity
            _, entity_keys, _ = gmsh_model.mesh.getKeysForElement(el, 'CurlHcurlLegendre0', returnCoord=False)



            # compute all products 1/det(dF)*(dF.w)^T v(B) (dF.w)
            for k in range(num_el_dofs):
                
                # the basis function in global coordinates
                B_ret[i, :] += J[0, :, :] @ curl_w_hat[3*k:3*(k+1)] * solution[entity_keys[k] - 1] / det_J[0]

        else:

            # mark this point as outside
            outside_mask[i] = True

    return B_ret